/*
*  qm_libraryview.cpp
*  QUIMUP library treeview
*  (c) 2008-2010 Johan Spee
*
*  This file is part of Quimup
*
*  QUIMUP is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  QUIMUP is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program. If not, see http://www.gnu.org/licenses/.
*/


#include "qm_libraryview.h"

qm_libview::qm_libview(QWidget *parent)
{
    setParent(parent);
    
    if (objectName().isEmpty())
        setObjectName("lib_view");
        
    config = NULL;
    mpd_connected = false;
    b_dragdrop = false;
    plist_droppos = 0;
    searched_mode_used = -1;
    b_panel_max = false;
    setRootIsDecorated(true);
    setItemsExpandable (true);
    setUniformRowHeights(true);
    setAllColumnsShowFocus(false);
    setSelectionMode(ExtendedSelection);
    sortByColumn (-1); // no sorting
    setSortingEnabled(false);
    setColumnCount(1);
    setDragDropMode(QAbstractItemView::DragOnly);
    setDropIndicatorShown(false);
    setDragEnabled(true);
    
    QObject::connect( this, SIGNAL( itemExpanded(QTreeWidgetItem *) ), this, SLOT( on_item_expanded(QTreeWidgetItem *)));
    QObject::connect( this, SIGNAL( itemSelectionChanged() ), this, SLOT( on_SelectionChanged()));
    
    // context menu
    ctxMenu = new QMenu();
    this->setContextMenuPolicy(Qt::DefaultContextMenu);
    
    QAction *append = new QAction(ctxMenu);
    append->setText(trUtf8("Append to playlist"));
    append->setIcon(QIcon(":/tr_plistappend.png"));
    QObject::connect( append, SIGNAL( triggered() ), this, SLOT( on_append() ) );
    ctxMenu->addAction(append);
    
    QAction *createnew = new QAction(ctxMenu);
    createnew->setText(trUtf8("Open in new playlist"));
    createnew->setIcon(QIcon(":/tr_plistnew.png"));
    QObject::connect( createnew, SIGNAL( triggered() ), this, SLOT( on_createnew() ) );
    ctxMenu->addAction(createnew);
    
    a_delete = new QAction(ctxMenu);
    a_delete->setText(trUtf8("Delete selected"));
    a_delete->setIcon(QIcon(":/tr_redcross.png"));
    QObject::connect( a_delete, SIGNAL( triggered() ), this, SLOT( on_delete() ) );
    ctxMenu->addAction(a_delete);
    
    a_opendir = new QAction(ctxMenu);
    a_opendir->setText(trUtf8("Open folder"));
    a_opendir->setIcon(QIcon(":/tr_folderopen.png"));
    QObject::connect( a_opendir, SIGNAL( triggered() ), this, SLOT( on_opendir() ) );
    ctxMenu->addAction(a_opendir);
    
    QAction * a_collapse = new QAction(ctxMenu);
    a_collapse->setText(trUtf8("Collapse all"));
    a_collapse->setIcon(QIcon(":/tr_collapse.png"));
    QObject::connect( a_collapse, SIGNAL( triggered() ), this, SLOT( collapseAll() ) );
    ctxMenu->addAction(a_collapse);
    
    a_opendir->setEnabled(false);
}


void qm_libview::on_SelectionChanged()
{
    QList<QTreeWidgetItem*> slist = selectedItems();
    
    if (slist.empty())
        return;
        
    QListIterator<QTreeWidgetItem *> itr(slist);
    while (itr.hasNext())
    { // no selection of non-draggable items
        QTreeWidgetItem *current = itr.next();
        if (current->gettype >= TP_PLISTITEM)
            current->setSelected(false);
        else
        { // no selection if if a parent is already selected
            bool a_parent_selected = false;
            QTreeWidgetItem * up = current->parent();
            while (up != NULL)
            {
                if ( up->isSelected() )
                    a_parent_selected = true;
                up = up->parent();
            }
            
            if (a_parent_selected)
                current->setSelected(false);
        }
    }
    slist.clear();
}


void qm_libview::on_item_expanded(QTreeWidgetItem * active_item)
{
    if (!mpd_connected)
        return;
        
    switch (active_item->gettype)
    {
    case TP_ARTIST:
        {
            if(active_item->getdummy)
            {
                take_dummy_from(active_item);
                expand_artist(active_item);
            }
            break;
        }
        
    case TP_ALBUM:
        {
            if(active_item->getdummy)
            {
                take_dummy_from(active_item);
                expand_album(active_item);
            }
            break;
        }
        
    case TP_GENRE:
        {
            if(active_item->getdummy)
            {
                take_dummy_from(active_item);
                expand_genre(active_item);
            }
            break;
        }
        
    case TP_PLAYLIST:
        {
            if(active_item->getdummy)
            {
                take_dummy_from(active_item);
                expand_playlist(active_item);
            }
            break;
        }
        
    case TP_FOLDER:
        {
            if(active_item->getdummy)
            {
                take_dummy_from(active_item);
                expand_folder(active_item);
            }
            break;
        }
        
    default: // TP_SONG, TP_PLISTITEM, TP_COMMENT
        break;
    }
}


void qm_libview::get_albums(QString searchfor, bool sort_by_date)
{
    clear();
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        headerItem()->setText(0, trUtf8(" not connected "));
        return;
    }
    
    qm_itemList itemlist;
    
    if (searchfor == "*^*")
        itemlist = mpdCom->get_albumlist(sort_by_date);
    else
        itemlist = mpdCom->get_album_searchlist(searchfor);
        
    if (itemlist.empty())
    {
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        headerItem()->setText(0, trUtf8(" no albums found "));
        return;
    }
    
    bool bydate = (sort_by_date &&  searchfor == "*^*");
    
    if (bydate)
        setColumnCount(3);
    else
        setColumnCount(2);
        
    int count = 0;
    
    itemlist.sort();
    // Skip duplicates!
    QString album = "*^*";
    QString artist = "*^*";
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (album == iter->album && artist == iter->artist)
            continue;
            
        album = iter->album;
        artist = iter->artist;
        
        new_item =  new QTreeWidgetItem();
        if (iter->album.isEmpty())
            new_item->setText(0, "?");
        else
            new_item->setText(0, iter->album);
        if (iter->artist.isEmpty())
            new_item->setText(1, "?");
        else
            new_item->setText(1, iter->artist);
        new_item->setIcon(0, QIcon(":/tr_album.png"));
        
        new_item->settype(TP_ALBUM);
        new_item->setartist(iter->artist);
        new_item->setalbum(iter->album);
        
        if (bydate)
            new_item->setText(2, iter->moddate);
            
        addTopLevelItem(new_item);
        add_dummy_to(new_item);
        count ++;
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" albums ");
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    headerItem()->setText(0, str);
    headerItem()->setText(1, trUtf8("artist") );
    if (bydate)
        headerItem()->setText(2, trUtf8("Last modified") );
    this->header()->resizeSection(0, 200);
    
    if (bydate)
        resizeColumnToContents(2);
        
    if (b_panel_max)
    {
        int w = viewport()->width();
        
        if (bydate)
        {
            resizeColumnToContents(0);
            resizeColumnToContents(1);
            resizeColumnToContents(2);
            
            if (w/3 < columnWidth(2))
                setColumnWidth(2, w/3);
                
            w = w - columnWidth(2);
            
            setColumnWidth(0, w/2 + 30);
            setColumnWidth(1, w/2 - 30);
        }
        else
        {
            setColumnWidth(0, w/2 + 30);
            setColumnWidth(1, w/2 - 30);
        }
    }
}
/////
void qm_libview::get_albums_year()
{

    clear();
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        headerItem()->setText(0, trUtf8(" not connected "));
        return;
    }
    
    qm_itemList itemlist;
    
    itemlist = mpdCom->get_yearlist();
        
    if (itemlist.empty())
    {
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        headerItem()->setText(0, trUtf8(" no albums found "));
        return;
    }
    
    setColumnCount(3);
        
    int count = 0;
    
    itemlist.sort();
	
    // Skip duplicates!
    QString album = "*^*";
    QString artist = "*^*";
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (album == iter->album && artist == iter->artist)
            continue;
            
        album = iter->album;
        artist = iter->artist;
        
        new_item =  new QTreeWidgetItem();
        if (iter->album.isEmpty())
            new_item->setText(0, "?");
        else
            new_item->setText(0, iter->album);
        if (iter->artist.isEmpty())
            new_item->setText(1, "?");
        else
            new_item->setText(1, iter->artist);
        new_item->setIcon(0, QIcon(":/tr_album.png"));
        
        new_item->settype(TP_ALBUM);
        new_item->setartist(iter->artist);
        new_item->setalbum(iter->album);
        
        new_item->setText(2, iter->year);
            
        addTopLevelItem(new_item);
        add_dummy_to(new_item);
        count ++;
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" albums ");
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    headerItem()->setText(0, str);
    headerItem()->setText(1, trUtf8("artist") );
	
    headerItem()->setText(2, trUtf8("year") );
    this->header()->resizeSection(0, 200);
    
    resizeColumnToContents(2);
        
    if (b_panel_max)
    {
        int w = viewport()->width();
        
            resizeColumnToContents(0);
            resizeColumnToContents(1);
            resizeColumnToContents(2);
            
            if (w/3 < columnWidth(2))
                setColumnWidth(2, w/3);
                
            w = w - columnWidth(2);
            
            setColumnWidth(0, w/2 + 30);
            setColumnWidth(1, w/2 - 30);
    }
}





// tree root items in folder mode
void qm_libview::get_folders()
{
    clear();
    
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setText(0, trUtf8(" not connected ") );
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        return;
    }
    
    qm_itemList itemlist;
    
    itemlist = mpdCom->get_folderlist("");
    
    if (itemlist.empty())
    {
        headerItem()->setText(0, trUtf8(" no folders found ") );
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        return;
    }
    
    itemlist.sort();
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if ( iter->type == TP_FOLDER )
        {
            new_item = new QTreeWidgetItem();
            new_item->setText(0, iter->name);
            new_item->setIcon(0, QIcon(":/tr_folder.png"));
            
            new_item->settype(TP_FOLDER);
            new_item->setfile(iter->file);
            new_item->setname(iter->name);
            
            addTopLevelItem(new_item);
            add_dummy_to(new_item);
        }
        
        if ( iter->type == TP_SONG )
        {
            new_item = new QTreeWidgetItem();
            new_item->setText(0, iter->name);
            new_item->setIcon(0, QIcon(":/tr_track.png"));
            add_songdata_toItem(new_item, *iter);
            addTopLevelItem(new_item);
        }
    }
    
    itemlist.clear();
    headerItem()->setText(0, trUtf8(" folders") );
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    
    if (b_panel_max)
    {
        int w = viewport()->width();
        setColumnWidth(0, w);
    }
}


// tree root items in palylist mode
void qm_libview::get_playlists()
{
    clear();
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setText(0, trUtf8(" not connected ") );
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        return;
    }
    
    qm_itemList itemlist = mpdCom->get_playlistlist();
    
    if (itemlist.empty())
    {
        headerItem()->setText(0, trUtf8(" no playlists found ") );
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        return;
    }
    
    itemlist.sort();
    
    int count = 0;
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (iter->name.isEmpty())
            iter->name = "?";
            
        new_item = new QTreeWidgetItem();
        new_item->setText(0, iter->name);
        new_item->setIcon(0, QIcon(":/tr_playlist.png"));
        
        new_item->settype(TP_PLAYLIST);
        new_item->setfile(iter->file);
        new_item->setname(iter->name);
        
        addTopLevelItem(new_item);
        add_dummy_to(new_item);
        
        count ++;
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" playlists ");
    headerItem()->setText(0, str);
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    resizeColumnToContents(0);
    
    if (b_panel_max)
    {
        int w = viewport()->width();
        setColumnWidth(0, w);
    }
}


// tree items in artist mode
void qm_libview::get_artists(QString searchfor)
{
    clear();
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setText(0, trUtf8(" not connected ") );
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        return;
    }
    
    qm_itemList itemlist;
    
    if (searchfor == "*^*")
        itemlist = mpdCom->get_artistlist();
    else
        itemlist = mpdCom->get_artist_searchlist(searchfor);
        
    if (itemlist.empty())
    {
        headerItem()->setText(0, trUtf8(" no artists found "));
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        return;
    }
    
    int count = 0;
    
    itemlist.sort();
    // skip duplicates!
    QString artist = "*^*";
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (artist == iter->artist)
            continue;
            
        artist = iter->artist;
        
        new_item = new QTreeWidgetItem();
        if (iter->artist.isEmpty())
            new_item->setText(0, "?");
        else
            new_item->setText(0, iter->artist);
        new_item->setIcon(0, QIcon(":/tr_artist.png"));
        
        new_item->settype(TP_ARTIST);
        new_item->setartist(iter->artist);
        new_item->setalbum(iter->album);
        new_item->setname(iter->name);
        
        addTopLevelItem(new_item);
        add_dummy_to(new_item);
        
        count ++;
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" artists ");
    headerItem()->setText(0, str);
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    
    if (b_panel_max)
    {
        int w = viewport()->width();
        setColumnWidth(0, w);
    }
}


void qm_libview::get_songs(QString searchfor)
{
    clear();
    
    if (!mpd_connected)
    {
        headerItem()->setText(0, trUtf8(" not connected ") );
        return;
    }
    
    qm_itemList itemlist = mpdCom->get_song_searchlist(searchfor);
    
    if (itemlist.empty())
    {
        headerItem()->setText(0, trUtf8(" no songs found ") );
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        return;
    }
    
    setColumnCount(2);
    int count = 0;
    itemlist.sort();
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        new_item = new QTreeWidgetItem();
        if (iter->title.isEmpty())
        {
            // use file name as title
            int i = (iter->file).lastIndexOf( '/' ) + 1;
            iter->title = (iter->file).right((iter->file).length() - i);
        }
        new_item->setText(0, iter->title);
        if (iter->artist.isEmpty())
            new_item->setText(1, "?");
        else
            new_item->setText(1, iter->artist);
            
        new_item->setIcon(0, QIcon(":/tr_track.png"));
        add_songdata_toItem(new_item, *iter);
        addTopLevelItem(new_item);
        count ++;
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" songs ");
    headerItem()->setText(0, str);
    headerItem()->setText(1, trUtf8(" artist") );
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    this->header()->resizeSection(0, 200);
    
    if (b_panel_max)
    {
        int w = viewport()->width();
        setColumnWidth(0, w/2 + 30);
        setColumnWidth(1, w/2 - 30);
    }
}

void qm_libview::get_genres(QString searchfor)
{
    clear();
    setColumnCount(1);
    
    if (!mpd_connected)
    {
        headerItem()->setText(0, trUtf8(" not connected ") );
        headerItem()->setIcon(0, QIcon(":/tr_redcross.png"));
        return;
    }
    
    qm_itemList itemlist = mpdCom->get_genrelist(searchfor);
    
    if (itemlist.empty())
    {
        headerItem()->setText(0, trUtf8(" no genres found ") );
        headerItem()->setIcon(0, QIcon(":/tr_dot.png"));
        return;
    }
    
    int count = 0;
    itemlist.sort();
    QTreeWidgetItem *new_item;
    QString genre = "*^*";
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if(iter->genre != genre)
        {
            genre = iter->genre;
            new_item = new QTreeWidgetItem();
            if (iter->genre.isEmpty())
                new_item->setText(0, "?");
            else
                new_item->setText(0, iter->genre);
            new_item->settype(TP_GENRE);
            new_item->setgenre(iter->genre);
            new_item->setIcon(0, QIcon(":/tr_genre.png"));
            addTopLevelItem(new_item);
            add_dummy_to(new_item);
            count++;
        }
    }
    itemlist.clear();
    
    QString str = " " + QString::number(count) + trUtf8(" genres ");
    headerItem()->setText(0, str);
    headerItem()->setIcon(0, QIcon(":/tr_check.png"));
    
    if (b_panel_max)
    {
        int w = viewport()->width();
        setColumnWidth(0, w);
    }
}


void qm_libview::expand_album(QTreeWidgetItem * active_item)
{
    qm_itemList itemlist;
    
    if ( config->browser_dbasemode == LB_ID_GENRE)
        itemlist = mpdCom->get_songlist(active_item->getalbum, active_item->getartist, active_item->getgenre);
    else
        itemlist = mpdCom->get_songlist(active_item->getalbum, active_item->getartist, "*^*");
        
        
    if (itemlist.empty())
        return;
        
    itemlist.sort();
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if ( iter->type == TP_SONG )
        {
            if (iter->title.isEmpty())
            {
                // use file name as title
                int i = (iter->file).lastIndexOf( '/' ) + 1;
                iter->title = (iter->file).right((iter->file).length() - i);
            }
            if (iter->artist.isEmpty())
                iter->artist = "?";
            new_item = new QTreeWidgetItem();
            new_item->setText(0, iter->track + " " + iter->title);
            new_item->setIcon(0, QIcon(":/tr_track.png"));
            add_songdata_toItem(new_item, *iter);
            active_item->addChild(new_item);
            addTopLevelItem(new_item);
        }
    }
    itemlist.clear();
}


void qm_libview::expand_artist(QTreeWidgetItem * active_item)
{

    qm_itemList itemlist = mpdCom->get_albums_for_artist(active_item->getartist);
    
    if (itemlist.empty())
        return;
        
    if (config->sort_albums_year)// add the date to the sorter
    {
        std::list<qm_listitemInfo>::iterator it;
        for (it = itemlist.begin(); it != itemlist.end(); ++it)
        {
            it->year  = mpdCom->get_album_year( it->album, it->artist);
            it->sorter = it->year + it->sorter;
        }
    }
    
    itemlist.sort();
    // skip duplicates!
    QString album = "*^*";
    
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (album == iter->album)
            continue;
            
        album = iter->album;
        
        new_item =  new QTreeWidgetItem();
        
        if (config->sort_albums_year)
        {
            if (iter->album.isEmpty())
                new_item->setText(0, iter->year + "?");
            else
                new_item->setText(0, iter->year + iter->album);
        }
        else
        {
            if (iter->album.isEmpty())
                new_item->setText(0, "?");
            else
                new_item->setText(0, iter->album);
        }
        
        new_item->setIcon(0, QIcon(":/tr_album.png"));
        
        new_item->settype(TP_ALBUM);
        new_item->setartist(iter->artist);
        new_item->setalbum(iter->album);
        new_item->setname(iter->name);
        
        active_item->addChild(new_item);
        add_dummy_to(new_item);
    }
    itemlist.clear();
}


void qm_libview::expand_playlist(QTreeWidgetItem * active_item)
{
    qm_itemList itemlist = mpdCom->get_playlist_itemlist(active_item->getfile);
    
    if (itemlist.empty())
        return;
        
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        new_item = new QTreeWidgetItem();
        new_item->setfile(iter->file);
        new_item->settype(TP_PLISTITEM);
        new_item->setForeground( 0, QBrush(QColor("#70707A")));
        if (iter->type == TP_STREAM)
        {
            new_item->setIcon(0, QIcon(":/tr_streamplayed.png"));
            new_item->setText(0, iter->file);
        }
        else
        {
            new_item->setIcon(0, QIcon(":/tr_trackplayed.png"));
            new_item->setText(0, iter->artist + " : " + iter->title);
            new_item->settime(iter->time);
            new_item->setartist(iter->artist);
            new_item->setalbum(iter->album);
            new_item->settitle(iter->title);
            new_item->settrack(iter->track);
            new_item->setname(iter->name);
            new_item->setgenre(iter->genre);
        }
        active_item->addChild(new_item);
    }
    itemlist.clear();
}

void qm_libview::expand_genre(QTreeWidgetItem * active_item)
{
    qm_itemList itemlist = mpdCom->get_genre_artistlist(active_item->getgenre);
    
    if (itemlist.empty())
        return;
        
    itemlist.sort();
    
    QTreeWidgetItem *artist_item = new QTreeWidgetItem();
    QString current_artist = "*^*";
    QTreeWidgetItem *album_item = new QTreeWidgetItem();
    QString current_album = "*^*";
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if (iter->artist != current_artist)
        {
            // add artist to active_item
            current_artist = iter->artist;
            artist_item = new QTreeWidgetItem();
            if (iter->artist.isEmpty())
                artist_item->setText(0, "?");
            else
                artist_item->setText(0, iter->artist);
            artist_item->setIcon(0, QIcon(":/tr_artist.png"));
            artist_item->settype(TP_ARTIST);
            artist_item->setartist(iter->artist);
            artist_item->setgenre(iter->genre);
            active_item->addChild(artist_item);
            
            // add albums to artist_item
            current_album = iter->album;
            album_item = new QTreeWidgetItem();
            if (iter->album.isEmpty())
                album_item->setText(0, "?");
            else
                album_item->setText(0, iter->album);
            album_item->setIcon(0, QIcon(":/tr_album.png"));
            album_item->settype(TP_ALBUM);
            album_item->setartist(iter->artist);
            album_item->setalbum(iter->album);
            album_item->setgenre(iter->genre);
            add_dummy_to(album_item);
            artist_item->addChild(album_item);
        }
        else // add more albums to artist_item
        {
            if (iter->album != current_album)
            {
                current_album = iter->album;
                album_item = new QTreeWidgetItem();
                if (iter->album.isEmpty())
                    album_item->setText(0, "?");
                else
                    album_item->setText(0, iter->album);
                album_item->setIcon(0, QIcon(":/tr_album.png"));
                album_item->settype(TP_ALBUM);
                album_item->setartist(iter->artist);
                album_item->setalbum(iter->album);
                album_item->setgenre(iter->genre);
                add_dummy_to(album_item);
                artist_item->addChild(album_item);
                ;
            }
        }
    }
    itemlist.clear();
}


void qm_libview::expand_folder(QTreeWidgetItem * active_item)
{
    qm_itemList itemlist;
    
    itemlist = mpdCom->get_folderlist(active_item->getfile);
    
    if (itemlist.empty())
        return;
    else
        itemlist.sort();
        
    QTreeWidgetItem *new_item;
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if ( iter->type == TP_FOLDER )
        {
            new_item = new QTreeWidgetItem();
            new_item->setText(0, iter->name);
            new_item->setIcon(0, QIcon(":/tr_folder.png"));
            
            new_item->settype(TP_FOLDER);
            new_item->setfile(iter->file);
            new_item->setname(iter->name);
            
            active_item->addChild(new_item);
            add_dummy_to(new_item);
        }
        
        if ( iter->type == TP_SONG )
        {
            new_item = new QTreeWidgetItem();
            new_item->setText(0, iter->name);
            new_item->setIcon(0, QIcon(":/tr_track.png"));
            add_songdata_toItem(new_item, *iter);
            active_item->addChild(new_item);
        }
    }
    
    itemlist.clear();
}

// from qm_browser::set_font
void qm_libview::set_font(QFont & thefont)
{
    setFont(thefont);
}


void qm_libview::on_mpd_plistsaved()
{
    if (config->browser_dbasemode == LB_ID_PLIST)
        reload_root();
}


void qm_libview::set_mpdCom(qm_mpdCom* conn, bool isconnected)
{
    mpd_connected = isconnected;
    mpdCom = conn;
    
    QObject::connect( mpdCom, SIGNAL( sgnl_plistsaved()), this, SLOT( on_mpd_plistsaved()));
    
    if (mpd_connected && config != NULL)
        reload_root();
        
    if (!config->mpd_rm_allowed)
        a_delete->setEnabled(false);
    else
        if(config->browser_dbasemode == LB_ID_PLIST)
            a_delete->setEnabled(true);
            
    if (mpd_connected)
        set_local_features(config->mpd_local_features);
    else
        set_local_features(false);
}

void qm_libview::contextMenuEvent ( QContextMenuEvent * event )
{
    if (event->reason() == QContextMenuEvent::Mouse && mpd_connected)
        ctxMenu->exec(QCursor::pos());
}


void qm_libview::add_songdata_toItem(QTreeWidgetItem *theItem, qm_listitemInfo data)
{
    theItem->settype(data.type);
    theItem->settime(data.time);
    theItem->setartist(data.artist);
    theItem->setalbum(data.album);
    theItem->settitle(data.title);
    theItem->setfile(data.file);
    theItem->settrack(data.track);
    theItem->setname(data.name);
    theItem->setgenre(data.genre);
}

void qm_libview::add_dummy_to(QTreeWidgetItem *theItem)
{
    QTreeWidgetItem *dummy = new QTreeWidgetItem();
    dummy->setText(0, "*^*");
    theItem->addChild(dummy);
    theItem->setdummy(true);
}

void qm_libview::take_dummy_from(QTreeWidgetItem *theItem)
{
    theItem->takeChild(0);
    theItem->setdummy(false);
}


void qm_libview::reload_root()
{
    if (!mpd_connected)
        return;
        
    a_delete->setEnabled(false);
    a_opendir->setEnabled(false);
    
    switch (config->browser_dbasemode)
    {
    case LB_ID_ARTIST:
        {
            get_artists("*^*");
            break;
        }
    case LB_ID_ALBUM:
        {
            get_albums("*^*", false);
            break;
        }
    case LB_ID_PLIST:
        {
            get_playlists();
            if (config->mpd_rm_allowed)
                a_delete->setEnabled(true);
            break;
        }
    case LB_ID_FOLDER:
        {
            get_folders();
            if (config->mpd_local_features)
                a_opendir->setEnabled(true);
            break;
        }
        
    case LB_ID_BYDATE:
        {
            get_albums("*^*", true);
            break;
        }
		
    case LB_ID_BYYEAR:
        {
            get_albums_year();
            break;
        }	
        
    case LB_ID_GENRE:
        {
            get_genres("*^*");
            break;
        }
        
    case LB_ID_SEARCH:
        {
            searched_mode_used = config->browser_searchmode;
            switch (searched_mode_used)
            {
            case SR_ID_ARTIST:
                {
                    get_artists(config->search_string);
                    break;
                }
            case SR_ID_ALBUM:
                {
                    get_albums(config->search_string);
                    break;
                }
            case SR_ID_TITLE:
                {
                    get_songs(config->search_string);
                    break;
                }
            case SR_ID_GENRE:
                {
                    get_genres(config->search_string);
                    break;
                }
            default:
                break;
            } // end switch searched_mode
            
            break;
        }
    default:
        break;
    }
}


void qm_libview::on_append()
{
    b_dragdrop = false;
    create_theModList(false);
}

void qm_libview::on_createnew()
{
    b_dragdrop = false;
    create_theModList(true);
    if (config->auto_playfirst && !thePlaylistModList->isEmpty())
        mpdCom->play(true);
}

void qm_libview::on_delete()
{
    if (config->browser_dbasemode != LB_ID_PLIST || !mpd_connected)
        return;
        
    QList<QTreeWidgetItem*> tobeRemoved = selectedItems();
    if (tobeRemoved.empty())
        return;
        
    // Ask for confirmation
    if ( QMessageBox::warning( this, trUtf8("Warning"),
                               trUtf8("This will permanently remove the \nselected items from your disk. "), trUtf8("Delete"), trUtf8("Cancel"), QString::null, 1, 0 ) )
        return;
        
    qm_commandList newCommandList;
    
    QListIterator<QTreeWidgetItem *> itr(tobeRemoved);
    itr.toBack();
    while (itr.hasPrevious())
    {
        QTreeWidgetItem *current = itr.previous();
        qm_mpd_command newCommand;
        newCommand.cmd = CMD_DPL;
        newCommand.file = current->getfile;
        newCommandList.push_back(newCommand);
    }
    
    mpdCom->execute_cmds( newCommandList, false );
    newCommandList.clear();
    tobeRemoved.clear();
    reload_root();
}


void qm_libview::on_opendir()
{
    if (currentItem() == NULL)
        return;
        
    QString dirpath = currentItem()->getfile;
    if (currentItem()->gettype != TP_FOLDER)
        dirpath = dirpath.left(dirpath.lastIndexOf("/"));
        
    QProcess *proc;
    QString full_path = config->MPD_musicpath + dirpath;
    proc->startDetached(config->file_manager, QStringList() << full_path);
}


void qm_libview::on_plistdrop()
{
    b_dragdrop = true;
    create_theModList(false);
}


void qm_libview::set_config(qm_config *conf)
{
    if (conf == NULL)
        return;
        
    config = conf;
}

/*  Pointer to thePlaylistModList in browserWin.
 plist_view and lib_view share this list for playlist
 modifications like drag&drop operations. */
void qm_libview::set_playlistModList(qm_commandList* theList)
{
    thePlaylistModList = theList;
}

void qm_libview::create_theModList(bool b_new_list)
{
    thePlaylistModList->clear();
    
    if (topLevelItemCount() == 0 || !mpd_connected)
        return;
        
    QList<QTreeWidgetItem*> tobeAdded = selectedItems();
    if (tobeAdded.empty())
        return;
        
    QListIterator<QTreeWidgetItem *> itr(tobeAdded);
    while (itr.hasNext())
    {
        QTreeWidgetItem *currentItem = itr.next();
        switch (currentItem->gettype)
        {
        case TP_STREAM:
        case TP_SONG:
            {
                qm_mpd_command newCommand;
                if (b_dragdrop)
                    newCommand.cmd = CMD_INS;
                else
                    newCommand.cmd = CMD_ADD;
                newCommand.moveTo = plist_droppos;
                newCommand.type = currentItem->gettype;
                newCommand.file = currentItem->getfile;
                newCommand.time = currentItem->gettime;
                newCommand.artist = currentItem->getartist;
                newCommand.album = currentItem->getalbum;
                newCommand.title = currentItem->gettitle;
                newCommand.track = currentItem->gettrack;
                if (b_dragdrop)
                    thePlaylistModList->push_front(newCommand);
                else
                    thePlaylistModList->push_back(newCommand);
                break;
            }
            
        case TP_ARTIST:
            {
                if ( config->browser_dbasemode == LB_ID_GENRE)
                    append_toModlist("*^*", currentItem->getartist, currentItem->getgenre);
                else
                    append_toModlist("*^*", currentItem->getartist, "*^*");
                break;
            }
            
        case TP_ALBUM:
            {
                if ( config->browser_dbasemode == LB_ID_GENRE)
                    append_toModlist(currentItem->getalbum, currentItem->getartist, currentItem->getgenre);
                else
                    append_toModlist(currentItem->getalbum, currentItem->getartist, "*^*");
                break;
            }
            
        case TP_GENRE:
            {
                append_toModlist("*^*", "*^*", currentItem->getgenre);
                break;
            }
            
        case TP_PLAYLIST:
            {
                append_toModlist_plist(currentItem->getfile);
                break;
            }
            
        case TP_FOLDER:
            {
                append_toModlist_folder(currentItem->getfile);
                break;
            }
            
        default:
            break;
        }
    }
    
    tobeAdded.clear();
    
    if (!thePlaylistModList->isEmpty())
    {
        if (b_new_list)
            mpdCom->clear_list();
            
        mpdCom->execute_cmds( *thePlaylistModList, b_new_list );
        
        /*  A new list is created through mpd which signals plist_view
         else signal plist_view (true: drag&drop, false: append) */
        if (!b_new_list)
            emit sgnl_playlist_modification(b_dragdrop);
            
        b_dragdrop = false;
        plist_droppos = 0;
    }
}


void qm_libview::append_toModlist(QString album,QString artist, QString genre)
{
    qm_itemList itemlist = mpdCom->get_songlist(album, artist, genre);
    
    if (itemlist.empty())
        return;
        
    itemlist.sort();
    
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        qm_mpd_command newCommand;
        if (b_dragdrop)
            newCommand.cmd = CMD_INS;
        else
            newCommand.cmd = CMD_ADD;
        newCommand.moveTo = plist_droppos;
        newCommand.type =  iter->type;
        newCommand.file =   iter->file;
        newCommand.time =   iter->time;
        newCommand.artist = iter->artist;
        newCommand.album =  iter->album;
        newCommand.title =  iter->title;
        newCommand.track =  iter->track;
        
        if (b_dragdrop)
            thePlaylistModList->push_front(newCommand);
        else
            thePlaylistModList->push_back(newCommand);
    }
    
    itemlist.clear();
}

// calls itself when a folder is found within a folder
void qm_libview::append_toModlist_folder(QString path)
{
    qm_itemList itemlist = mpdCom->get_folderlist(path);
    itemlist.sort();
    
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        if ( iter->type == TP_FOLDER )
            append_toModlist_folder(iter->file);
        else
            if ( iter->type == TP_SONG  || iter->type == TP_STREAM )
            {
                qm_mpd_command newCommand;
                if (b_dragdrop)
                    newCommand.cmd = CMD_INS;
                else
                    newCommand.cmd = CMD_ADD;
                newCommand.moveTo = plist_droppos;
                newCommand.type =   iter->type;
                newCommand.file =   iter->file;
                newCommand.time =   iter->time;
                newCommand.artist = iter->artist;
                newCommand.album =  iter->album;
                newCommand.title =  iter->title;
                newCommand.track =  iter->track;
                if (b_dragdrop)
                    thePlaylistModList->push_front(newCommand);
                else
                    thePlaylistModList->push_back(newCommand);
            }
    }
    itemlist.clear();
}


void qm_libview::append_toModlist_plist(QString playlist)
{
    qm_itemList itemlist = mpdCom->get_playlist_itemlist(playlist);
    
    if (itemlist.empty())
        return;
        
    std::list<qm_listitemInfo>::iterator iter;
    for (iter = itemlist.begin(); iter != itemlist.end(); ++iter)
    {
        qm_mpd_command newCommand;
        if (b_dragdrop)
            newCommand.cmd = CMD_INS;
        else
            newCommand.cmd = CMD_ADD;
        newCommand.file = iter->file;
        newCommand.moveTo = plist_droppos;
        // special treatment for streams
        if (iter->type == TP_STREAM)
        {
            newCommand.artist = "Stream ";
            newCommand.time = 0;
            newCommand.album = "Info in player ";
            newCommand.track = "";
            newCommand.type = TP_STREAM;
            if (b_dragdrop)
                thePlaylistModList->push_front(newCommand);
            else
                thePlaylistModList->push_back(newCommand);
        }
        else
        {
            newCommand.time = iter->time;
            newCommand.type = TP_SONG;
            newCommand.artist = iter->artist;
            newCommand.album = iter->album;
            newCommand.title = iter->title;
            newCommand.track = iter->track;
            if (b_dragdrop)
                thePlaylistModList->push_front(newCommand);
            else
                thePlaylistModList->push_back(newCommand);
        }
    }
}


void qm_libview::on_plist_dropreceived(int pos)
{
    if (pos == -1) // drop on empty list
    {
        b_dragdrop = false;
        plist_droppos = 0;
        create_theModList(true);
    }
    else // drop on existing list
    {
        b_dragdrop = true;
        plist_droppos = pos;
        create_theModList(false);
    }
}


void qm_libview::startDrag(Qt::DropActions da)
{
    QMimeData *data;
    QModelIndexList indexes = selectedIndexes();
    if (indexes.count() > 0)
    {
        data = model()->mimeData(indexes);
        if (data == 0)
            return;
    }
    else
        return;
    QDrag* drag = new QDrag(this);
    drag->setPixmap(QPixmap(":/tr_dragdrop.png"));
    drag->setMimeData(data);
    drag->exec(da, Qt::IgnoreAction);
}


void qm_libview::set_local_features(bool enabled)
{
    if (config->browser_dbasemode == LB_ID_FOLDER )
        a_opendir->setEnabled(enabled);
}

void qm_libview::set_maxpanel(bool ismax)
{
    b_panel_max = ismax;
}


qm_libview::~qm_libview()
{}
