/*
 *  qm_mpdCom.cpp
 *  QUIMUP mpd communicator class
 *  (c) 2008-2010 Johan Spee
 *
 *  This file is part of Quimup
 *
 *  QUIMUP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QUIMUP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/.
 */

#include "qm_mpdcom.h"

qm_mpdCom::qm_mpdCom()
{
    if (objectName().isEmpty())
        setObjectName("mpdCom");
        
    conn = NULL;
    b_reload = true;
    b_mpdconf_found = false;
    b_connecting = false;
    b_statCheckBusy = false;
    b_dbaseUpdating = false;
    b_no_volume = false;
    mpdconf_path = "";
    plist_lenght = 0;
    current_playlist = -1;
    current_songID = -1;
    current_songpos = -1;
    current_status = 100;
    current_volume = 50;
    statusLoop = new QTimer(this);
    connect(statusLoop, SIGNAL(timeout()), this, SLOT(statusCheck()));
    reconnectLoop = new QTimer(this);
    connect(reconnectLoop, SIGNAL(timeout()), this, SLOT(mpd_reconnect()));
}

// set a pointer to the parent's config
void qm_mpdCom::set_config(qm_config *conf)
{
    config = conf;
}

void qm_mpdCom::configure()
{
    QString problem = "";
    autodetect = false;
    b_mpdconf_found = false;
    
    switch (config->profile)
    {
    case 0:
        {
            autodetect = true;
        }
    case 1:
        {
            serverPort = config->MPD_port_1;
            serverName = config->MPD_host_1;
            serverPassword = config->MPD_password_1;
            mpdconf_path = config->MPD_cofigpath_1;
            break;
        }
    case 2:
        {
            serverPort = config->MPD_port_2;
            serverName = config->MPD_host_2;
            serverPassword = config->MPD_password_2;
            mpdconf_path = config->MPD_cofigpath_2;
            break;
        }
    case 3:
        {
            serverPort = config->MPD_port_3;
            serverName = config->MPD_host_3;
            serverPassword = config->MPD_password_3;
            mpdconf_path = config->MPD_cofigpath_3;
            break;
        }
    case 4:
        {
            serverPort = config->MPD_port_4;
            serverName = config->MPD_host_4;
            serverPassword = config->MPD_password_4;
            mpdconf_path = config->MPD_cofigpath_4;
            break;
        }
    default:
        break;
        // just in case...
        serverPassword = serverPassword.left(serverPassword.lastIndexOf("@"));
    }
    
    
    if (autodetect)
    {
        int config_count = 0;
        QString thispath;
        QString allConfigs;
        QFile file;
        
        printf ("Auto-detecting MPD's config file.... ");
        QString homeDir = QDir::homePath();
        if (!homeDir.endsWith("/"))
            homeDir.append("/");
            
        thispath = "/etc/mpd.conf";
        file.setFileName(thispath);
        if ( file.exists() )
        {
            mpdconf_path = thispath;
            printf (" " + mpdconf_path.toUtf8() + " ");
            b_mpdconf_found = true;
            allConfigs = thispath;
            config_count++;
        }
        
        thispath = homeDir + ".mpd/mpd.conf";
        file.setFileName(thispath);
        if ( file.exists() )
        {
            mpdconf_path = thispath;
            printf (" " + mpdconf_path.toUtf8() + " ");
            b_mpdconf_found = true;
            allConfigs.append("\n"+ thispath);
            config_count++;
        }
        
        thispath = homeDir + ".mpdconf";
        file.setFileName(thispath);
        if ( file.exists() )
        {
            mpdconf_path = thispath;
            printf (" " + mpdconf_path.toUtf8() + " ");
            b_mpdconf_found = true;
            allConfigs.append("\n"+ thispath);
            config_count++;
        }
        
        
        if (b_mpdconf_found)
        {
            if (config_count > 1)
            {
                printf ("\nMultiple MPD config files found! Using: " + mpdconf_path.toUtf8() + "\n");
                show_messagebox(trUtf8("Multiple MPD config files found!"),
                                allConfigs + "\n\n" +
                                trUtf8("The default choice is ") + mpdconf_path +
                                trUtf8("\n\nPlease create a login profile with the file that MPD is actually using.") );
            }
            else
                printf ("\nusing: " + mpdconf_path.toUtf8() + "\n");
                
            config->MPD_cofigpath_0 = mpdconf_path;
        }
        else
        {
            problem = trUtf8("No mpd config file was not found. Is mpd installed?");
            mpdconf_path = "";
            printf ("not found!\nIs mpd installed?\n");
            config->MPD_cofigpath_0 = "?";
        }
    }
    else // not autodetect
    {
        if (!mpdconf_path.isEmpty())
        {
            QFile file(mpdconf_path);
            if (file.exists())
            {
                b_mpdconf_found = true;
                printf ("MPD config file exists\n");
            }
            else
            {
                problem = trUtf8("Configured mpd config file does not exist");
                printf ("Configured mpd config file does not exist!\n");
                b_mpdconf_found =  false;
            }
        }
        else
            printf ("No configured mpd config file! Some features will not be available.\n");
    }
    
    
    if (autodetect)
    {
        if (b_mpdconf_found)
        {
            serverPort = get_port();
            serverName = get_host();
            if (serverName.isEmpty())
                serverName = "localhost";
            serverPassword = get_password();
        }
        else
        {
            serverPort = 6600;
            serverName = "localhost";
            serverPassword = "";
        }
        config->MPD_port_0 = serverPort;
        config->MPD_host_0 = serverName;
        config->MPD_password_0 = serverPassword;
    }
    
    QString musicpath = "";
    if (b_mpdconf_found)
    {
        musicpath = get_musicPath();
        if (musicpath.isEmpty())
        {
            problem = trUtf8("Could not determine mpd's music path");
            printf ("Music path not found in mpd's config file\n");
        }
        else
        {
            printf ("Music path found in mpd's config file\n");
            QDir dir(musicpath);
            if (!dir.exists())
            {
                problem = trUtf8("Music path found in mpd's config, but it does not exist");
                printf ("...but it does not exist!\n");
                musicpath = "";
            }
        }
    }
    
    config->MPD_musicpath = musicpath;
    
    if (musicpath.isEmpty())
        config->mpd_local_features = false;
    else
        config->mpd_local_features = true;
        
    if (!problem.isEmpty())
        show_messagebox(problem, trUtf8("Check the 'connect' tab of the settings window."));
}


void qm_mpdCom::setOutputs(qm_outputList outputlist)
{
    if (conn == NULL)
        return;
        
    bool wasplaying = false;
    if (current_status == MPD_STATE_PLAY)
    {
        wasplaying = true;
        pause();
    }
    
    QListIterator<qm_output> i(outputlist);
    while (i.hasNext())
    {
        qm_output output = i.next();
        if (output.enabled)
            mpd_send_enable_output(conn, output.id);
        else
            mpd_send_disable_output(conn, output.id);
        mpd_response_finish(conn);
        errorCheck("setOutputs");
    }
    
    if (wasplaying)
        resume();
}


qm_outputList qm_mpdCom::getOutputs()
{
    qm_outputList outputlist;
    // get current output devices
    if (conn == NULL)
        return outputlist;
        
    mpd_output *output;
    mpd_send_outputs(conn);
    if (errorCheck("getOutputs"))
    {
        while ((output = mpd_recv_output(conn)))
        {
            qm_output qoutput;
            qoutput.id  = mpd_output_get_id(output);
            qoutput.enabled  = mpd_output_get_enabled(output);
            qoutput.name = QString::fromUtf8(mpd_output_get_name(output));
            outputlist.push_back(qoutput);
        }
        
    }
    mpd_response_finish(conn);
    
    if (output != NULL)
        mpd_output_free(output);
        
    return outputlist;
}


QString qm_mpdCom::get_musicPath()
{
    QString result = get_string("music_directory");
    if (!result.isEmpty())
    {
        if (!result.endsWith("/"))
            result.append("/");
    }
    return result;
}


QString qm_mpdCom::get_playlistPath()
{
    QString result = get_string("playlist_directory");
    if (!result.isEmpty())
    {
        if (!result.endsWith("/"))
            result.append("/");
    }
    return result;
}


QString qm_mpdCom::get_host()
{
    QString result = get_string("bind_to_address");
    return result;
}


void qm_mpdCom::updatefile(QString file)
{
    if (conn == NULL)
        return;
        
    if (config->mpd_rescan_allowed)
        mpd_send_rescan(conn, (const char*)file.toUtf8());
    else
        if (config->mpd_update_allowed)
            mpd_send_update(conn, (const char*)file.toUtf8());
        else
            return;
            
    mpd_response_finish(conn);
    if (errorCheck("updatefile"))
    {
		current_playlist++; // prevent list update
        current_songID = -1; //trigger reloading the song
        printf ("MPD database update: OK\n");
    }
}


int qm_mpdCom::get_port()
{
    QString get = get_string("port")
                  ;
    return get.toInt();
}


QString qm_mpdCom::get_password()
{
    QString result = get_string("password");
    result = result.left(result.lastIndexOf("@"));
    return result;
}


QString qm_mpdCom::get_string(QString key)
{
    if (!b_mpdconf_found)
        return "";
        
    QFile file(mpdconf_path);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return "";
        
    QTextStream in(&file);
    while (!in.atEnd())
    {
        QString line = in.readLine();
        if (line.contains(key, Qt::CaseInsensitive) && !line.startsWith("#"))
        {
            QString result = line.section('"', 1, 1);
            return result.simplified();
            break;
        }
    }
    
    file.close();
    return "";
}


void qm_mpdCom::update_all()
{
    if (conn == NULL)
    {
        b_dbaseUpdating = false;
        emit sgnl_update_ready();
        return;
    }
    b_dbaseUpdating = TRUE;
    mpd_send_update(conn, (char*)"/");
    mpd_response_finish(conn);
    errorCheck("update_all");
    // when ready: see statusloop
}


bool qm_mpdCom::mpd_connect()
{
    // stop the status loop
    statusLoop->stop();
    // disconnect if connected
    if (conn != NULL)
    {
        statusLoop->stop();
        emit sgnl_connected(false);
        current_playlist = -1;
        current_songID = -1;
        current_songpos = -1;
        current_status = -1;
        mpd_connection_free(conn);
        conn = NULL;
    }
    
    b_connecting = true;
    
    // no server, no go.
    if (serverName.isEmpty() && autodetect)
    {
        serverName = "localhost";
        printf ("No host name: using \"localhost\"\n");
    }
    
    // port number must be > 1024
    if (!(serverPort > 1024))
    {
        printf ("Invalid port: %i [using 6600]\n", serverPort);
        serverPort = 6600;
    }
    
    conn = mpd_connection_new( serverName.toUtf8(), serverPort ,30000);
    
    if ( mpd_connection_get_error(conn) != MPD_ERROR_SUCCESS )
    {
        mpd_connection_clear_error(conn);
        if (conn != NULL)
        {
            mpd_connection_free(conn);
            conn = NULL;
        }
    }
    
    if (conn != NULL)
        mpd_response_finish(conn);
        
        
    // If connection failed try 4 more tmes at 0.2 sec intervals
    // (in case mpd is starting up)
    int count = 1;
    while (conn == NULL && count < 5)
    {
        // wait 200 msec
        usleep(200000);
        
        // try again
        conn = mpd_connection_new( serverName.toUtf8(), serverPort ,30000);
        if ( mpd_connection_get_error(conn) != MPD_ERROR_SUCCESS )
        {
            mpd_connection_clear_error(conn);
            if (conn != NULL)
            {
                mpd_connection_free(conn);
                conn = NULL;
            }
        }
        
        if (conn != NULL)
            mpd_response_finish(conn);
            
        count++;
    }
    
    if (conn == NULL) // Still no luck
    {
        printf ("Connection failed (tried %i times)\nIs MPD running?\n", count);
        b_connecting = false;
        emit sgnl_connected(false);
        return false;
    }
    else
    {
        printf ("Connected to MPD on try %i\n", count);
    }
    
    
    if (!serverPassword.isEmpty())
    {
        mpd_send_password(conn, serverPassword.toUtf8());
        mpd_response_finish(conn);
        if ( mpd_connection_get_error(conn) != MPD_ERROR_SUCCESS )
        {
            printf ("Error: %s\n", mpd_connection_get_error_message(conn));
            mpd_connection_clear_error(conn);
            show_messagebox(trUtf8("Password for mpd was refused "), trUtf8( "Check out the 'connect' tab in the 'settings' window."));
        }
        else
            printf("Password accepted\n");
    }
    
    // Let's tickle MPD
    char *command;
    mpd_pair *mp;
    bool b_status_allowed = false;
    config->mpd_outputs_allowed = false;
    config->mpd_update_allowed = false;
    config->mpd_deleteid_allowed = false;
    config->mpd_save_allowed = false;
    config->mpd_single_allowed = false;
    config->mpd_consume_allowed = false;
    config->mpd_xfade_allowed = false;
    config->mpd_repeat_allowed = false;
    config->mpd_random_allowed = false;
    config->mpd_rescan_allowed = false;
    config->mpd_shuffle_allowed = false;
    config->mpd_rm_allowed = false;
    
    mpd_send_allowed_commands (conn);
    while ( (mp = mpd_recv_command_pair(conn)) != NULL)
    {
        command = (char*)mp->value;
        //printf("command: %s\n", command);
        if (!strcmp(command, "status"))
            b_status_allowed = true;
        if (!strcmp(command, "enableoutput"))
            config->mpd_outputs_allowed = true;
        if (!strcmp(command, "update"))
            config->mpd_update_allowed = true;
        if (!strcmp(command, "deleteid"))
            config->mpd_deleteid_allowed = true;
        if (!strcmp(command, "rm"))
            config->mpd_rm_allowed = true;
        if (!strcmp(command, "save"))
            config->mpd_save_allowed = true;
        if (!strcmp(command, "single"))
            config->mpd_single_allowed = true;
        if (!strcmp(command, "consume"))
            config->mpd_consume_allowed = true;
        if (!strcmp(command, "crossfade"))
            config->mpd_xfade_allowed = true;
        if (!strcmp(command, "repeat"))
            config->mpd_repeat_allowed = true;
        if (!strcmp(command, "rescan"))
            config->mpd_rescan_allowed = true;
        if (!strcmp(command, "random"))
            config->mpd_random_allowed = true;
        if (!strcmp(command, "shuffle"))
            config->mpd_shuffle_allowed = true;
        mpd_return_pair(conn, mp);
    }
    
    if (!b_status_allowed)
    {
        mpd_connection_free(conn);
        conn = NULL;
        b_connecting = false;
        emit sgnl_connected(false);
        printf ("No permission to get mpd's status (Password required?). Connection closed.\n");
        show_messagebox(trUtf8("No permission to get mpd's status"), trUtf8("Possibly a password is required. Check out the 'connect' tab in the 'settings' window."));
        return false;
    }
    // we're in business
    b_connecting = false;
    
    emit sgnl_connected(true);
    
    statusCheck();
    
    // poll mpd every 200 msec
    statusLoop->start(200);
    
    return true;
}


bool qm_mpdCom::mpd_reconnect()
{
    // previous loop has not finished
    if (b_connecting)
        return true;
        
    // try to connect
    b_connecting = true;
    
    conn = mpd_connection_new( serverName.toUtf8(), serverPort ,30000);
    
    if ( mpd_connection_get_error(conn) != MPD_ERROR_SUCCESS )
    {
        mpd_connection_clear_error(conn);
        if (conn != NULL)
        {
            mpd_connection_free(conn);
            conn = NULL;
        }
    }
    
    if (conn == NULL)
    {
        b_connecting = false;
        return true;
    }
    
    mpd_response_finish(conn);
    
    // login if there is a password
    if (!serverPassword.isEmpty())
    {
        mpd_send_password(conn, serverPassword.toUtf8());
        mpd_response_finish(conn);
    }
    
    // connected!
    printf ("Reconnected\n");
    reconnectLoop->stop();
    b_connecting = false;
    emit sgnl_connected(true);
    statusCheck();
    // poll mpd every 200 msec
    statusLoop->start(200);
    return false;
}


void qm_mpdCom::prev()
{
    if (conn == NULL)
        return;
    mpd_send_previous(conn);
    mpd_response_finish(conn);
    errorCheck("mpd_sendPrevCommand");
}

void qm_mpdCom::stop()
{
    if (conn == NULL)
        return;
    mpd_send_stop(conn);
    mpd_response_finish(conn);
    errorCheck("mpd_sendStopCommand");
}

void qm_mpdCom::play(bool reload)
{
    if (conn == NULL)
        return;
    if (reload)
        b_reload = true;
    mpd_send_play(conn);
    mpd_response_finish(conn);
    errorCheck("mpd_sendPlayCommand");
}

void qm_mpdCom::play_this(int ID)
{
    if (conn == NULL)
        return;
    mpd_send_play_id(conn, ID);
    mpd_response_finish(conn);
    errorCheck("mpd_sendPlayIdCommand");
}

void qm_mpdCom::resume()
{
    if (conn == NULL)
        return;
    mpd_send_pause(conn, 0);
    mpd_response_finish(conn);
    errorCheck("mpd_sendPauseCommand_0");
}

void qm_mpdCom::pause()
{
    if (conn == NULL)
        return;
    mpd_send_pause(conn, 1);
    mpd_response_finish(conn);
    errorCheck("mpd_sendPauseCommand_1");
}

void qm_mpdCom::next()
{
    if (conn == NULL)
        return;
    mpd_send_next(conn);
    mpd_response_finish(conn);
    errorCheck("mpd_sendNextCommand");
}

void qm_mpdCom::set_random(bool status)
{
    if (conn == NULL)
        return;
    mpd_send_random(conn, status);
    mpd_response_finish(conn);
    errorCheck("set_random");
}

void qm_mpdCom::set_repeat(bool status)
{
    if (conn == NULL)
        return;
    mpd_send_repeat(conn, status);
    mpd_response_finish(conn);
    errorCheck("set_repeat");
}


void qm_mpdCom::set_single(bool status)
{
    if (conn == NULL)
        return;
    mpd_send_single(conn, status);
    mpd_response_finish(conn);
    errorCheck("set_single");
}


void qm_mpdCom::set_consume(bool status)
{
    if (conn == NULL)
        return;
    mpd_send_consume(conn, status);
    mpd_response_finish(conn);
    errorCheck("set_consume");
}

void qm_mpdCom::volume_up(int i)
{
    if (conn == NULL || b_no_volume)
        return;
    int vol = current_volume + i;
    if (vol > 100)
        vol = 100;
    mpd_send_set_volume(conn, vol);
    mpd_response_finish(conn);
    errorCheck("volume_up");
}

void qm_mpdCom::volume_down(int i)
{
    if (conn == NULL || b_no_volume)
        return;
    int vol = current_volume - i;
    if (vol < 0)
        vol = 0;
    mpd_send_set_volume(conn, vol);
    mpd_response_finish(conn);
    errorCheck("volume_down");
}

void qm_mpdCom::set_volume(int vol)
{
    if (conn == NULL || b_no_volume)
        return;
    mpd_send_set_volume(conn, vol);
    mpd_response_finish(conn);
    errorCheck("set_volume");
}

void qm_mpdCom::set_seek(int percent)
{
    if (conn == NULL)
        return;
    mpd_send_seek_pos(conn, current_songpos, percent);
    mpd_response_finish(conn);
    errorCheck("set_seek");
}


bool qm_mpdCom::errorCheck(QString action)
{
    if (conn == NULL)
    {
        statusLoop->stop();
        emit sgnl_connected(false);
        printf ("Error: on \"" + action.toUtf8() + "\": %s\n", mpd_connection_get_error_message(conn));
        printf ("Reconnecting...\n");
        current_playlist = -1;
        plist_lenght = 0;
        current_songID = -1;
        current_songpos = -1;
        current_status = -1;
        mpd_reconnect();
        // still no luck: try to reconnect every 2 seconds
        if (conn == NULL)
            reconnectLoop->start(2000);
        return false;
    }
    
    if (mpd_connection_get_error(conn) != MPD_ERROR_SUCCESS)
    {
    
        // If it is a non-fatal error (no permission etc.).
        if (mpd_connection_get_error(conn) != MPD_ERROR_CLOSED)
        {
            printf ("Non fatal error: on \"" + action.toUtf8() + "\": %s\n", mpd_connection_get_error_message(conn));
            mpd_connection_clear_error(conn);
            mpd_response_finish(conn);
            return true;
        }
        
        // A fatal error occurred: disconnnect & reconnect
        printf ("Fatal error: on \"" + action.toUtf8() + "\": %s\n", mpd_connection_get_error_message(conn));
        statusLoop->stop();
        mpd_connection_clear_error(conn);
        mpd_response_finish(conn);
        mpd_connection_free(conn);
        current_playlist = -1;
        printf ("Reconnecting...\n");
        plist_lenght = 0;
        current_songID = -1;
        current_songpos = -1;
        current_status = -1;
        conn = NULL;      // first this
        emit sgnl_connected(false);  // next this
        mpd_reconnect();
        // still no luck: reconnect every 2 seconds
        if (conn == NULL)
            reconnectLoop->start(2000);
            
        return false;
    }
    
    // No problems
    mpd_connection_clear_error(conn);
    return true;
}


void qm_mpdCom::set_xfade(int secs)
{
    if (conn == NULL)
        return ;
    mpd_send_crossfade(conn, secs);
    mpd_response_finish(conn);
    errorCheck("set_xfade");
}


int qm_mpdCom::get_xfade()
{
    if (conn == NULL)
        return 0;
        
    mpd_send_status(conn);
    mpd_status * servStatus = mpd_recv_status(conn);
    mpd_response_finish(conn);
    
    if (!errorCheck("get_xfade") || servStatus == NULL)
    {
        if (servStatus != NULL)
        {
            mpd_status_free(servStatus);
            mpd_response_finish(conn);
        }
        return 0;
    }
    else
    {
        int i = mpd_status_get_crossfade(servStatus);
        mpd_status_free(servStatus);
        mpd_response_finish(conn);
        return i;
    }
}


bool qm_mpdCom::statusCheck()
{
    // return "false" terminates the loop
    if (!errorCheck("Starting statusCheck"))
        return false;
        
    // "true" keeps the loop going
    if (b_statCheckBusy)
        return true;
        
    b_statCheckBusy = true;
    
    // Get the current server status
    mpd_send_status(conn);
    
    mpd_status * servStatus = mpd_recv_status(conn);
    
    mpd_response_finish(conn);
    
    if (!errorCheck("mpd_recv_status") || servStatus == NULL)
    {
        b_statCheckBusy = false;
        if (servStatus != NULL && conn != NULL)
        {
            mpd_status_free(servStatus);
            mpd_response_finish(conn);
        }
        return true;
    }
    
    // server has not found a supported mixer
    if (mpd_status_get_volume(servStatus) == -1)
    {
        if (!b_no_volume)
        {
            printf ("MPD: No supported mixer found\n");
            b_no_volume = true;
        }
    }
    else
        b_no_volume = false;
        
    // database update was requested && has finished
    if (b_dbaseUpdating && mpd_status_get_update_id(servStatus) == 0)
    {
        get_statistics();
        b_dbaseUpdating = false;
        emit sgnl_update_ready();
    }
    
    // playlist was updated
    if (current_playlist < (int)mpd_status_get_queue_version(servStatus) )
    {
        printf ("New playlist\n");
        current_playlist = mpd_status_get_queue_version(servStatus);
        get_playlist();
        // reload the song (streams use this to update song info)
        b_reload = true;
    }
    
    struct mpd_song *song = NULL;
    // the song changed, or new playlist
    if( (current_songID != mpd_status_get_song_id(servStatus))  || b_reload)
    {
        b_reload = false;
        // get the new song.
        mpd_entity * entity = NULL;
        mpd_send_current_song(conn);
        entity = mpd_recv_entity(conn);
        mpd_response_finish(conn);
        
        // empty list, last has finished or first hasn't started
        if (entity == NULL || !errorCheck("statuscheck: mpd_send_current_song"))
        {
            current_songpos = -1;
            current_songID = -1;
            song = NULL;
        }
        else
        {
            current_songpos = mpd_status_get_song_pos(servStatus);
            current_songID = mpd_status_get_song_id(servStatus);
            song = mpd_song_dup (mpd_entity_get_song(entity));
        }
        
        // free memory
        if (entity != NULL)
            mpd_entity_free(entity);
            
        // collect & send the song info
        qm_songInfo newSongInfo = get_songInfo_from(song);
        emit sgnl_songInfo(newSongInfo);
    }
    
    // send the status info
    emit sgnl_statusInfo(servStatus);
    
    current_status = mpd_status_get_state(servStatus);
    current_volume = mpd_status_get_volume(servStatus);
    
    // free memory
    if (servStatus != NULL)
        mpd_status_free(servStatus);
    if (song != NULL)
        mpd_song_free (song);
        
    // done
    b_statCheckBusy = false;
    return true;
}

qm_songInfo qm_mpdCom::get_songInfo_from(mpd_song *theSong)
{
    qm_songInfo newSongInfo;
    
    if (theSong == NULL)
        newSongInfo.nosong = true;
    else
    {
        newSongInfo.nosong = false;
        
        newSongInfo.file = QString::fromUtf8(mpd_song_get_uri(theSong));
        if ( newSongInfo.file.startsWith("http:"))
        {
            newSongInfo.type = TP_STREAM;
            newSongInfo.time = 0;
            
        }
        else
        {
            newSongInfo.type = TP_SONG;
            newSongInfo.time = mpd_song_get_duration(theSong);
        }
        
        newSongInfo.songID  = mpd_song_get_id(theSong);
        newSongInfo.songNr  = mpd_song_get_pos(theSong);
        
        newSongInfo.name  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_NAME, 0));
        newSongInfo.artist  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_ARTIST, 0));
        newSongInfo.title  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_TITLE, 0));
        newSongInfo.album  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_ALBUM, 0));
        newSongInfo.track  = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_TRACK, 0)) );
        newSongInfo.disc  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_DISC, 0));
        newSongInfo.date  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_DATE, 0));
        // convert xxxx-xx-xx format to xxxx
        newSongInfo.date = newSongInfo.date.left(4);
        newSongInfo.genre  = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_GENRE, 0));
        newSongInfo.comment = QString::fromUtf8(mpd_song_get_tag(theSong, MPD_TAG_COMMENT, 0));
    }
    return newSongInfo;
}

QString qm_mpdCom::get_version()
{
    QString version = "";
    if (conn == NULL)
        return version;
        
    version += QString::number(mpd_connection_get_server_version(conn)[0]) + ".";
    version += QString::number(mpd_connection_get_server_version(conn)[1]) ;
    
    if ( mpd_connection_get_server_version(conn)[0] == 0 && mpd_connection_get_server_version(conn)[1] < 16)
        show_messagebox( trUtf8("Connected to MPD ") + version, trUtf8("\nMPD 0.16 or above is required. \nSome functions will not work! ") );
        
    return version;
}


void qm_mpdCom::get_playlist()
{
    qm_songList newPlayList;
    plist_lenght = 0;
    
    if (conn == NULL)
    {
        emit sgnl_plistupdate(newPlayList, -10);
        return;
    }
    
    struct mpd_song *song = NULL;
    mpd_entity * entity;
    mpd_send_list_queue_meta(conn);
    while((entity = mpd_recv_entity(conn)))
    {
        if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG)
        {
            song = mpd_song_dup (mpd_entity_get_song(entity));
            qm_songInfo newSongInfo = get_songInfo_from(song);
            newPlayList.push_back(newSongInfo);
            plist_lenght++;
        }
        mpd_entity_free(entity);
    }
    mpd_response_finish(conn);
    errorCheck("get_playlist");
    if (song != NULL)
        mpd_song_free (song);
        
    int changedID = get_listChanger();
    // signal the library window
    emit sgnl_plistupdate(newPlayList, changedID);
    newPlayList.clear();
}


// cmdlist must start with the highest songnr!
void qm_mpdCom::execute_cmds(qm_commandList cmdlist, bool resetlist)
{
    if(cmdlist.empty() ||conn == NULL)
        return;
        
    mpd_command_list_begin(conn, true);
    
    QListIterator<qm_mpd_command> itr(cmdlist);
    while (itr.hasNext())
    {
        qm_mpd_command curCommand = itr.next();
        
        switch (curCommand.cmd)
        {
        case CMD_DPL: // delete stored playlist
            {
                mpd_send_rm(conn, curCommand.file.toUtf8() );
                break;
            }
            
        case CMD_ADD:
            {
                plist_lenght++;
                current_playlist++; // prevent list update
                mpd_send_add(conn, curCommand.file.toUtf8());
                break;
            }
            
        case CMD_INS:
            {
                // add
                plist_lenght++;
                current_playlist++; // prevent list update
                mpd_send_add(conn, curCommand.file.toUtf8());
                // move into position
                current_playlist++; // prevent list update
                mpd_send_move(conn, plist_lenght-1, curCommand.moveTo);
                break;
            }
            
        case CMD_MOV: // move ID to pos
            {
                current_playlist++; // prevent list update
                mpd_send_move_id(conn, curCommand.songid, curCommand.moveTo);
                break;
            }
            
        case CMD_POS: // move from pos to pos
            {
                current_playlist++; // prevent list update
                mpd_send_move(conn, curCommand.songnr, curCommand.moveTo);
                break;
            }
            
        case CMD_DEL:
            {
                plist_lenght--;
                current_playlist++; // prevent list update
                mpd_send_delete_id(conn, curCommand.songid);
                break;
            }
            
        default:
            break;
        } // end switch
    }
    mpd_command_list_end(conn);
    mpd_response_finish(conn);
    errorCheck("execute_cmds");
    
    cmdlist.clear();
    
    current_songID = -1;
    current_songpos = -1;
    current_status = -1;
    
    if (resetlist)
        current_playlist = -1;
}

void qm_mpdCom::clear_list()
{
    if (conn == NULL)
        return;
    mpd_send_clear(conn);
    mpd_response_finish(conn);
    errorCheck("clear_list");
    current_songID = -1;
    current_songpos = -1;
    current_status = -1;
    current_playlist = -1;
}


void qm_mpdCom::shuffle_list()
{
    if (plist_lenght < 2 || conn == NULL)
        return;
        
    mpd_run_shuffle(conn);
    mpd_response_finish(conn);
    errorCheck("shuffle_list_1");
    
    // move current song to the top
    if (current_songID != -1)
    {
        current_playlist++; // prevent reloading
        mpd_send_move_id(conn, current_songID, 0);
        mpd_response_finish(conn);
        errorCheck("shuffle_list_2");
    }
    // trigger reloading the list
    current_playlist = -1;
}


void qm_mpdCom::get_statistics()
{
    if (conn == NULL)
    {
        emit sgnl_dbase_statistics("Library data not available");
        return;
    }
    
    mpd_stats *sData;
    mpd_send_stats(conn);
    sData = mpd_recv_stats(conn);
    mpd_response_finish(conn);
    if (!errorCheck("mpd_recv_stats") || sData == NULL)
    {
        emit sgnl_dbase_statistics("Library data not available");
        if (sData != NULL)
            mpd_stats_free(sData);
        return;
    }
    
    QString statistics =  "<i>Artists</i> " + QString::number(mpd_stats_get_number_of_artists(sData)) +
                          " - <i>Albums</i> "  + QString::number(mpd_stats_get_number_of_albums(sData)) +
                          " - <i>Songs</i> "  + QString::number(mpd_stats_get_number_of_songs(sData)) ;
                          
    emit sgnl_dbase_statistics(statistics);
    
    if (sData != NULL)
        mpd_stats_free(sData);
}


QString qm_mpdCom::save_playlist(QString listname)
{
    if (conn == NULL)
    {
        return "Error: Mpd is not connected.";
    }
    
    mpd_send_save(conn, listname.toUtf8() );
    mpd_response_finish(conn);
    
    emit sgnl_plistsaved();
    
    if (!errorCheck("save_playlist"))
    {
        return "Error: " + QString::fromUtf8(mpd_connection_get_error_message(conn));
    }
    else
        return "Playlist \"" + listname + "\" was successfully saved.";
        
        
}


QString qm_mpdCom::save_selection(QStringList tobeSaved, QString listname)
{
    if (conn == NULL)
    {
        return "Error: Mpd is not connected.";
    }
    
    mpd_command_list_begin(conn, true);
    for (int i = 0; i < tobeSaved.size(); ++i)
    {
        mpd_send_playlist_add(conn, listname.toUtf8(), tobeSaved.at(i).toUtf8());
    }
    mpd_command_list_end(conn);
    mpd_response_finish(conn);
    
    emit sgnl_plistsaved();
    
    if (!errorCheck("save_selection"))
    {
        return "Error: " + QString::fromUtf8(mpd_connection_get_error_message(conn));
    }
    else
        return "Playlist \"" + listname + "\" was successfully saved.";
}


void qm_mpdCom::show_messagebox(QString message, QString info)
{
    QMessageBox msgBox;
    msgBox.setWindowTitle(trUtf8("Problem"));
    msgBox.setIcon(QMessageBox::Warning);
    msgBox.setText("<b>" + message + " </b>");
    msgBox.setInformativeText(info);
    msgBox.exec();
}


qm_itemList qm_mpdCom::get_yearlist()
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString year = "*^*";
    QString newyear;
    QString album = "*^*";
    QString newalbum;
	
    struct mpd_song *song;
    mpd_send_list_all_meta(conn, "/");
    while ((song = mpd_recv_song (conn)))
    {
        if (mpd_song_get_tag (song, MPD_TAG_DATE, 0) != NULL )
            newyear = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_DATE, 0));
        else
            newyear = "";
			
		// convert xxx-xx-xx format toi xxxx
		if (!newyear.isEmpty())
        newyear = newyear.left(4) + " ";
            
        if (mpd_song_get_tag (song, MPD_TAG_ALBUM, 0) != NULL )
        {
            newalbum = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ALBUM, 0));
            
            if (album != newalbum || year != newyear)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
				newItem.artist = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ARTIST, 0));
				if (newyear.isEmpty())
				{
					newItem.sorter = "0000" + newalbum.toLower() + newItem.artist.toLower();
					newItem.year = "?";
				}
				else
				{
					newItem.sorter = newyear + newalbum.toLower() + newItem.artist.toLower();
					newItem.year = newyear;
				}
                itemlist.push_back(newItem);
                album = newalbum;
                year = newyear;
            }
        }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_yearlist");
    
    return itemlist;
}

qm_itemList qm_mpdCom::get_albumlist(bool sort_by_date)
{
    qm_itemList itemlist;

	
    if (conn == NULL)
        return itemlist;
		
    int nowtime = (QDateTime::currentDateTime()).toTime_t();
	int nowtime_length = QString::number(nowtime).length(); 
	
    QString album = "*^*";
    QString newalbum;
    QString artist = "*^*";
    QString newartist;
    struct mpd_song *song;
    mpd_send_list_all_meta(conn, "/");
    while ((song = mpd_recv_song (conn)))
    {
        if (mpd_song_get_tag (song, MPD_TAG_ARTIST, 0) != NULL )
            newartist = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ARTIST, 0));
        else
            newartist = "";
            
        if (mpd_song_get_tag (song, MPD_TAG_ALBUM, 0) != NULL )
        {
            newalbum = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ALBUM, 0));
            
            if (album != newalbum || artist != newartist)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
                newItem.artist = newartist;
                newItem.sorter = newalbum.toLower() + newartist.toLower();
                if (sort_by_date)
                {
                    QDateTime time;
                    time.setTime_t(mpd_song_get_last_modified(song));
                    newItem.moddate = time.toString("yy/MM/dd  hh:mm");
					// invert the sort order
					
					int oldlast = nowtime - time.toTime_t();
					QString sort  = QString::number( oldlast );
					
					while (sort.length() < nowtime_length)
						sort = "0" + sort;
					
                    newItem.sorter = sort + newItem.sorter;
                }
                itemlist.push_back(newItem);
                album = newalbum;
                artist = newartist;
            }
        }
        
        if (mpd_song_get_tag (song, MPD_TAG_ALBUM, 0) == NULL)
        {
            newalbum = "";
            
            if (artist != newartist)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
                newItem.artist = newartist;
                newItem.sorter = newalbum.toLower() + newartist.toLower();
                if (sort_by_date)
                {
                    QDateTime time;
                    time.setTime_t(mpd_song_get_last_modified(song));
                    newItem.moddate = time.toString("yy/MM/dd  hh:mm");
					// invert the sort order
					
					int oldlast = nowtime - time.toTime_t();
					QString sort  = QString::number( oldlast );
					
					while (sort.length() < nowtime_length)
						sort = "0" + sort;
					
                    newItem.sorter = sort + newItem.sorter;
                }
                itemlist.push_back(newItem);
                album = newalbum;
                artist = newartist;
            }
        }
        
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_albumlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_album_searchlist(QString searchfor)
{
    if (searchfor == "*^*")
        return get_albumlist(false);
        
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString album = "*^*";
    QString newalbum;
    QString artist = "*^*";
    QString newartist;
    struct mpd_song *song;
    mpd_search_db_songs(conn, false);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ALBUM, searchfor.toUtf8());
    mpd_search_commit(conn);
    
    while ((song = mpd_recv_song (conn)))
    {
        if (mpd_song_get_tag (song, MPD_TAG_ARTIST, 0) != NULL )
            newartist = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ARTIST, 0));
        else
            newartist = "";
            
        if (mpd_song_get_tag (song, MPD_TAG_ALBUM, 0) != NULL )
        {
            newalbum = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ALBUM, 0));
            
            if (album != newalbum || artist != newartist)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
                newItem.artist = newartist;
                newItem.sorter = newalbum.toLower() + newartist.toLower();
                itemlist.push_back(newItem);
                album = newalbum;
                artist = newartist;
            }
        }
        else
            if (searchfor.isEmpty())
            {
                newalbum = "";
                
                if (!album.isEmpty() || artist != newartist)
                {
                    qm_listitemInfo newItem;
                    newItem.type = TP_ALBUM;
                    newItem.album = newalbum;
                    newItem.artist = newartist;
                    newItem.sorter = newalbum.toLower() + newartist.toLower();
                    itemlist.push_back(newItem);
                    album = newalbum;
                    artist = newartist;
                }
            }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_album_searchlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_albums_for_artist(QString artist)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString album = "*^*";
    QString newalbum;
    struct mpd_song *song;
    mpd_search_db_songs(conn, true);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ARTIST, artist.toUtf8());
    mpd_search_commit(conn);
    
    while ((song = mpd_recv_song (conn)))
    {
    
        if (mpd_song_get_tag (song, MPD_TAG_ALBUM, 0) != NULL )
        {
            newalbum = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ALBUM, 0));
            
            if (album != newalbum)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
                newItem.artist = artist;
                newItem.sorter = newalbum.toLower();
                itemlist.push_back(newItem);
                album = newalbum;
            }
        }
        else
        {
            newalbum = "";
            if (!album.isEmpty())
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ALBUM;
                newItem.album = newalbum;
                newItem.artist = artist;
                newItem.sorter = newalbum;
                itemlist.push_back(newItem);
                album = newalbum;
            }
        }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_album_searchlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_artistlist()
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString artist = "*^*";
    QString newartist;
    struct mpd_song *song;
    mpd_send_list_all_meta(conn, "/");
    while ((song = mpd_recv_song (conn)))
    {
    
        if (mpd_song_get_tag (song, MPD_TAG_ARTIST, 0) != NULL )
        {
            newartist = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ARTIST, 0));
            
            if (artist != newartist)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ARTIST;
                newItem.artist = newartist;
                newItem.sorter = newartist.toLower();
                if (config->ignore_leading_the && (newItem.sorter).startsWith("the ") )
                    newItem.sorter = (newItem.sorter).right((newItem.sorter).size() - 4);
                itemlist.push_back(newItem);
                artist = newartist;
            }
        }
        else
        {
            newartist = "";
            
            if (!artist.isEmpty())
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ARTIST;
                newItem.artist = newartist;
                newItem.sorter = newartist.toLower();
                if (config->ignore_leading_the && (newItem.sorter).startsWith("the ") )
                    newItem.sorter = (newItem.sorter).right((newItem.sorter).size() - 4);
                itemlist.push_back(newItem);
                artist = newartist;
            }
        }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_artistlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_artist_searchlist(QString searchfor)
{
    if (searchfor == "*^*")
        return get_artistlist();
        
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString artist = "*^*";
    QString newartist;
    struct mpd_song *song;
    mpd_search_db_songs(conn, false);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ARTIST, searchfor.toUtf8());
    mpd_search_commit(conn);
    
    while ((song = mpd_recv_song (conn)))
    {
    
        if (mpd_song_get_tag (song, MPD_TAG_ARTIST, 0) != NULL )
        {
            newartist = QString::fromUtf8(mpd_song_get_tag (song, MPD_TAG_ARTIST, 0));
            
            if (artist != newartist)
            {
                qm_listitemInfo newItem;
                newItem.type = TP_ARTIST;
                newItem.artist = newartist;
                newItem.sorter = newartist.toLower();
                if (config->ignore_leading_the && (newItem.sorter).startsWith("the ") )
                    newItem.sorter = (newItem.sorter).right((newItem.sorter).size() - 4);
                itemlist.push_back(newItem);
                artist = newartist;
            }
        }
        else
            if (searchfor.isEmpty())
            {
                newartist = "";
                
                if (!artist.isEmpty())
                {
                    qm_listitemInfo newItem;
                    newItem.type = TP_ARTIST;
                    newItem.artist = newartist;
                    newItem.sorter = newartist.toLower();
                    if (config->ignore_leading_the && (newItem.sorter).startsWith("the ") )
                        newItem.sorter = (newItem.sorter).right((newItem.sorter).size() - 4);
                    itemlist.push_back(newItem);
                    artist = newartist;
                }
            }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_album_searchlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_playlistlist()
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    struct mpd_entity *ent;
    mpd_send_list_meta (conn, "/");
    while ((ent = mpd_recv_entity (conn)))
    {
        if (mpd_entity_get_type (ent) == MPD_ENTITY_TYPE_PLAYLIST)
        {
            qm_listitemInfo newItem;
            newItem.type = TP_PLAYLIST;
            const struct mpd_playlist *playlist = mpd_entity_get_playlist(ent);
            QString path = QString::fromUtf8(mpd_playlist_get_path (playlist));
            newItem.file = path;
            // put only the list name in 'name'
            newItem.name = path.section("/",-1);
            newItem.sorter = newItem.name.toLower();
            itemlist.push_back(newItem);
        }
        mpd_entity_free (ent);
    }
    mpd_response_finish (conn);
    errorCheck("get_playlistlist");
    return itemlist;
}


qm_itemList qm_mpdCom::get_folderlist(QString filePath)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    struct mpd_entity * entity;
    mpd_send_list_meta(conn, filePath.toUtf8());
    while( (entity = mpd_recv_entity(conn)) != NULL )
    {
        if( mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_DIRECTORY )
        {
            qm_listitemInfo newItem;
            const struct mpd_directory * directory = mpd_entity_get_directory(entity);
            QString path = QString::fromUtf8(mpd_directory_get_path (directory));
            newItem.file = path;
            newItem.name = path.section("/",-1);
            newItem.type = TP_FOLDER;
            newItem.sorter = "0" + newItem.name.toLower(); // "0" so dirs come first
            
            itemlist.push_back(newItem);
        }
        else
            if(mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG)
            {
                qm_listitemInfo newItem;
                const struct mpd_song * song = mpd_entity_get_song (entity);
                newItem.file = QString::fromUtf8(mpd_song_get_uri(song));
                newItem.type = TP_SONG; // never a stream here
                newItem.time = mpd_song_get_duration(song);
                newItem.name = (newItem.file).section("/",-1);
                newItem.sorter = "1" + newItem.name.toLower(); // "1" so files come next
                newItem.artist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
                newItem.album = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
                newItem.title = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TITLE, 0));
                newItem.track = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TRACK, 0)) );
                newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
                itemlist.push_back(newItem);
            }
        mpd_entity_free(entity);
    } // end while
    mpd_response_finish (conn);
    errorCheck("get_folderlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_songlist(QString album, QString artist, QString genre)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    struct mpd_song *song;
    mpd_search_db_songs(conn, true);
    if (artist != "*^*")
        mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ARTIST, artist.toUtf8());
    if (album != "*^*")
        mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ALBUM, album.toUtf8());
    if (genre != "*^*")
        mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_GENRE, genre.toUtf8());
    mpd_search_commit(conn);
    
    while ( (song = mpd_recv_song(conn)) )
    {
        qm_listitemInfo newItem;
        newItem.file = QString::fromUtf8(mpd_song_get_uri(song));
        newItem.type = TP_SONG; // never a stream here
        newItem.time = mpd_song_get_duration(song);
        newItem.artist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
        newItem.album = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
        newItem.title = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TITLE, 0));
        newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
        
        if (mpd_song_get_tag(song, MPD_TAG_TRACK, 0))
        {
            newItem.track = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TRACK, 0)) );
            newItem.sorter = (newItem.artist + newItem.album + newItem.track + newItem.title).toLower();
        }
        else
            newItem.sorter = (newItem.artist + newItem.album + "00" + newItem.title).toLower();
        if (newItem.title.isEmpty())
            newItem.sorter = newItem.sorter + newItem.file;
            
        itemlist.push_back(newItem);
        mpd_song_free(song);
    }
    mpd_response_finish(conn);
    errorCheck("get_songlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_song_searchlist(QString searchfor)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    struct mpd_song *song;
    mpd_search_db_songs(conn, false);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_TITLE, searchfor.toUtf8());
    mpd_search_commit(conn);
    
    while ((song = mpd_recv_song (conn)))
    {
    
        if (mpd_song_get_tag (song, MPD_TAG_TITLE, 0) != NULL )
        {
            qm_listitemInfo newItem;
            newItem.file = QString::fromUtf8(mpd_song_get_uri(song));
            newItem.type = TP_SONG; // never a stream here
            newItem.time = mpd_song_get_duration(song);
            newItem.artist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
            newItem.album = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
            newItem.title = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TITLE, 0));
            newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
            
            if (mpd_song_get_tag(song, MPD_TAG_TRACK, 0))
            {
                newItem.track = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TRACK, 0)) );
                newItem.sorter = (newItem.artist + newItem.album + newItem.track + newItem.title).toLower();
            }
            else
                newItem.sorter = (newItem.artist + newItem.album + "00" + newItem.title).toLower();
            if (newItem.title.isEmpty())
                newItem.sorter = newItem.sorter + newItem.file;
                
            itemlist.push_back(newItem);
        }
        else
            if (searchfor.isEmpty())
            {
                qm_listitemInfo newItem;
                newItem.file = QString::fromUtf8(mpd_song_get_uri(song));
                newItem.type = TP_SONG; // never a stream here
                newItem.time = mpd_song_get_duration(song);
                newItem.artist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
                newItem.album = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
                newItem.title = "";
                newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
                
                if (mpd_song_get_tag(song, MPD_TAG_TRACK, 0))
                {
                    newItem.track = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TRACK, 0)) );
                    newItem.sorter = (newItem.artist + newItem.album + newItem.track + newItem.title).toLower();
                }
                else
                    newItem.sorter = (newItem.artist + newItem.album + "00" + newItem.title).toLower();
                if (newItem.title.isEmpty())
                    newItem.sorter = newItem.sorter + newItem.file;
                    
                itemlist.push_back(newItem);
            }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_song_searchlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_genre_artistlist(QString searchfor)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    QString artist = "*^*";
    QString newartist;
    QString album = "*^*";
    QString newalbum;
    struct mpd_song *song;
    mpd_search_db_songs(conn, true);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_GENRE, searchfor.toUtf8());
    mpd_search_commit(conn);
    
    while ((song = mpd_recv_song (conn)))
    {
        if ( mpd_song_get_tag (song, MPD_TAG_GENRE, 0) == searchfor )
        {
            newalbum = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
            newartist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
            if (artist != newartist || album != newalbum)
            {
                qm_listitemInfo newItem;
                newItem.genre = searchfor;
                newItem.type = TP_ARTIST;
                newItem.artist = newartist;
                newItem.album = newalbum;
                newItem.sorter = (newItem.artist + newItem.album).toLower();
                itemlist.push_back(newItem);
                artist =newartist;
                album = newalbum;
            }
        }
        mpd_song_free (song);
    }
    mpd_response_finish (conn);
    errorCheck("get_genre_artistlist");
    
    return itemlist;
}


qm_itemList qm_mpdCom::get_genrelist(QString searchfor)
{
    qm_itemList itemlist;
    if (conn == NULL)
        return itemlist;
        
    if (searchfor == "*^*") // get full list
    {
        QString newgenre;
        struct mpd_pair * pair;
        mpd_search_db_tags(conn, MPD_TAG_GENRE);
        mpd_search_commit(conn);
        
        while ((pair = mpd_recv_pair_tag (conn, MPD_TAG_GENRE)))
        {
            newgenre = QString::fromUtf8(pair->value);
            qm_listitemInfo newItem;
            newItem.genre = newgenre;
            newItem.sorter = newgenre.toLower();
            itemlist.push_back(newItem);
        }
        mpd_response_finish (conn);
        errorCheck("get_genrelist full");
    }
    else
    {
        struct mpd_song *song;
        mpd_search_db_songs(conn, false);
        mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_GENRE, searchfor.toUtf8());
        mpd_search_commit(conn);
        
        while ((song = mpd_recv_song (conn)))
        {
        
            if (mpd_song_get_tag (song, MPD_TAG_GENRE, 0) != NULL )
            {
                qm_listitemInfo newItem;
                newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
                newItem.sorter = newItem.genre; // not toLower() or libview mays show things like: Jazz jazz Jazz
                itemlist.push_back(newItem);
            }
            else
                if (searchfor.isEmpty())
                {
                    qm_listitemInfo newItem;
                    newItem.genre = "";
                    newItem.sorter = "0";
                    itemlist.push_back(newItem);
                }
            mpd_song_free (song);
        }
        mpd_response_finish (conn);
        errorCheck("get_genrelist search");
    }
    
    return itemlist;
}

QString qm_mpdCom::fixTrackformat(QString instring)
{
    // remove the x/xx format
    QString str = instring;
    str =  str.left(str.lastIndexOf("/"));
    // add leading zero, when required
    if (str.size() == 1)
        str = "0" + str;
    return str;
}


QString qm_mpdCom::get_album_year(QString album, QString artist)
{
    QString year = "";
    
    if (conn == NULL)
        return year;
        
    if (album.isEmpty() || artist.isEmpty())
        return year;
        
    struct mpd_pair * pair;
    mpd_search_db_tags(conn, MPD_TAG_DATE);
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ARTIST, artist.toUtf8());
    mpd_search_add_tag_constraint (conn, MPD_OPERATOR_DEFAULT, MPD_TAG_ALBUM,  album.toUtf8());
    mpd_search_commit(conn);
    
    while ((pair = mpd_recv_pair_tag (conn, MPD_TAG_DATE)))
    {
        // take the first that comes along
        year = QString::fromUtf8(pair->value);
        if (!year.isEmpty())
            break;
    }
    mpd_response_finish (conn);
    errorCheck("get_album_year");
    
    // convert xxx-xx-xx format toi xxxx
    if (!year.isEmpty())
        year = year.left(4) + " ";
        
    return year;
}


qm_IDlist qm_mpdCom::get_IDlist()
{
    qm_IDlist idlist;
    
    if (conn == NULL)
        return idlist;
        
    struct mpd_song *song = NULL;
    mpd_entity * entity;
    mpd_send_list_queue_meta(conn);
    while((entity = mpd_recv_entity(conn)))
    {
        if (mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG)
        {
            song = mpd_song_dup (mpd_entity_get_song(entity));
            idlist.push_back(mpd_song_get_id(song));
        }
        mpd_entity_free(entity);
    }
    mpd_response_finish(conn);
    errorCheck("mpd_send_list_queue_meta");
    if (song != NULL)
        mpd_song_free (song);
        
    return idlist;
}


void qm_mpdCom::reset_current_song()
{
    current_songID = -1;
}


qm_itemList qm_mpdCom::get_playlist_itemlist(QString playlist)
{
    qm_itemList itemlist;
    
    if (conn == NULL)
        return itemlist;
        
    struct mpd_song *song;
    mpd_send_list_playlist_meta(conn, playlist.toUtf8());
    
    while ((song = mpd_recv_song(conn)))
    {
        qm_listitemInfo newItem;
        newItem.file = QString::fromUtf8(mpd_song_get_uri(song));
        if (newItem.file.startsWith("http://"))
        {
            newItem.type = TP_STREAM;
            newItem.time = 0;
        }
        else
        {
            newItem.type = TP_SONG;
            newItem.time = mpd_song_get_duration(song);
        }
        newItem.artist = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ARTIST, 0));
        newItem.album = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_ALBUM, 0));
        newItem.title = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TITLE, 0));
        newItem.genre = QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_GENRE, 0));
        newItem.track = fixTrackformat(QString::fromUtf8(mpd_song_get_tag(song, MPD_TAG_TRACK, 0)) );
        
        itemlist.push_back(newItem);
        mpd_song_free (song);
    }
    mpd_response_finish(conn);
    errorCheck("get_playlist_itemlist");
    return itemlist;
}

// find out which song (ID) caused the playlist to update
int qm_mpdCom::get_listChanger()
{
    int id = -1;
    if (conn == NULL)
        return id;
        
    mpd_entity *entity = NULL;
    mpd_send_queue_changes_meta(conn, current_playlist-1);
    while (( entity = mpd_recv_entity(conn)) != NULL)
    {
        // get the first that comes along
        if(mpd_entity_get_type(entity) == MPD_ENTITY_TYPE_SONG)
            id = mpd_song_get_id( mpd_entity_get_song(entity) );
        mpd_entity_free(entity);
        if (id != -1)
            break;
    }
    mpd_response_finish(conn);
    errorCheck("get_listChanger");
    return id;
}

bool qm_mpdCom::getReplayGainMode(QString mode)
{
    bool b_ok = false;
    
    if (conn == NULL)
        return b_ok;
        
    if (!mpd_send_command(conn, "replay_gain_status", NULL))
        return b_ok;
        
    if (mpd_pair *pair = mpd_recv_pair_named(conn, "replay_gain_mode"))
    {
        QString result = pair->value;
        
        result.toLower();
        
        if (result == mode)
            b_ok =  true;
    }
    mpd_response_finish(conn);
    return b_ok;
}

void qm_mpdCom::setReplayGainMode(int mode)
{
    if (conn == NULL)
        return;
    const char *rg_mode;
    
    switch (mode)
    {
    case 0:
        rg_mode = "off";
        break;
    case 1:
        rg_mode = "track";
        break;
    case 2:
        rg_mode = "album";
        break;
    case 3:
        rg_mode = "auto";
        break;
    default:
        return;
    }
    mpd_send_command(conn, "replay_gain_mode", rg_mode, NULL);
    mpd_response_finish(conn);
}

qm_mpdCom::~qm_mpdCom()
{}
