/*
*  qm_playlistview.cpp
*  QUIMUP playlist listview
*  (c) 2008-2010 Johan Spee
*
*  This file is part of Quimup
*
*  QUIMUP is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  QUIMUP is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program. If not, see http://www.gnu.org/licenses/.
*/


#include "qm_playlistview.h"

qm_plistview::qm_plistview(QWidget *parent)
{
    setParent(parent);

    if (objectName().isEmpty())
        setObjectName("plist_view");

    // initVars
    dropBeforeIndex = -1;
    line = new QWidget(viewport());
    line->hide();
    line->setAutoFillBackground(true);
    QPalette pal = line->palette();
    pal.setColor(line->backgroundRole(), QColor("#5f8ba9"));
    line->setPalette(pal);
    setDragDropMode(QAbstractItemView::DragDrop);
    setDropIndicatorShown(false);
    setDragEnabled(true);

    current_songPos = -1;
    current_songID = -1;
    current_status = 0;
    plist_totaltime = 0;
    plist_droppos = 0;
    b_panel_max = false;
    b_streamplaying = false;
    mpd_connected = false;
    col_default_fg = this->palette().color(QPalette::Foreground);
    col_played_fg  = QColor("#70707A");
    col_playing_fg = QColor("#003070");

    //setupUI
    setRootIsDecorated(false);
    setItemsExpandable (false);
    setUniformRowHeights(true);
    setAllColumnsShowFocus(true);
    setSelectionMode(ExtendedSelection);
    sortByColumn (-1); // no sorting
    setColumnCount(6);
    QStringList labels;
    labels << "#" << "" << trUtf8(" artist ") << trUtf8(" title ") << trUtf8(" m:s ") << trUtf8(" album : nr ");
    setHeaderLabels(labels);
    headerItem()->setIcon(1, QIcon(":/tr_trackplayed.png"));
    QObject::connect( this, SIGNAL( itemDoubleClicked (QTreeWidgetItem *, int)), this, SLOT( on_item_dclicked(QTreeWidgetItem *)));

    ctxMenu = new QMenu();
    this->setContextMenuPolicy(Qt::DefaultContextMenu);

    QAction *clearlist = new QAction(ctxMenu);
    clearlist->setText(trUtf8("Clear list"));
    clearlist->setIcon(QIcon(":/tr_redcross.png"));
    QObject::connect( clearlist, SIGNAL( triggered() ), this, SLOT( clear_it() ) );
    ctxMenu->addAction(clearlist);

    a_delsel = new QAction(ctxMenu);
    a_delsel->setText(trUtf8("Remove selected"));
    a_delsel->setIcon(QIcon(":/tr_delete.png"));
    QObject::connect( a_delsel, SIGNAL( triggered() ), this, SLOT( delete_it() ) );
    ctxMenu->addAction(a_delsel);

    QAction *purge = new QAction(ctxMenu);
    purge->setText(trUtf8("Purge played"));
    purge->setIcon(QIcon(":/tr_purge.png"));
    QObject::connect( purge, SIGNAL( triggered() ), this, SLOT( purge_it() ) );
    ctxMenu->addAction(purge);

    a_shuffle = new QAction(ctxMenu);
    a_shuffle->setText(trUtf8("Shuffle list"));
    a_shuffle->setIcon(QIcon(":/tr_shuffle.png"));
    QObject::connect( a_shuffle, SIGNAL( triggered() ), this, SLOT( shuffle_it() ) );
    ctxMenu->addAction(a_shuffle);

    a_savelist = new QAction(ctxMenu);
    a_savelist->setText(trUtf8("Save list"));
    a_savelist->setIcon(QIcon(":/tr_save.png"));
    QObject::connect( a_savelist, SIGNAL( triggered() ), this, SLOT( save_it() ) );
    ctxMenu->addAction(a_savelist);

    a_savesel = new QAction(ctxMenu);
    a_savesel->setText(trUtf8("Save selected"));
    a_savesel->setIcon(QIcon(":/tr_saveselect.png"));
    QObject::connect( a_savesel, SIGNAL( triggered() ), this, SLOT( save_selected() ) );
    ctxMenu->addAction(a_savesel);
    ctxMenu->addSeparator();
}

void qm_plistview::reSize_columns()
{
    resizeColumnToContents(0);
    resizeColumnToContents(1);
    resizeColumnToContents(4);
    resizeColumnToContents(5);

    if (b_panel_max)
    {
        int w = viewport()->width();

        w = w - (columnWidth(0) + columnWidth(1) + columnWidth(4));
        setColumnWidth(2, w/3);
        setColumnWidth(3, w/3);
        setColumnWidth(5, w/3);
    }
}

void qm_plistview::set_maxpanel(bool ismax)
{
    b_panel_max = ismax;
}

void qm_plistview::reNum_reID_list()
{
    qm_IDlist idlist = mpdCom->get_IDlist();

    if (idlist.length() != topLevelItemCount())
        mpdCom->get_playlist();

    int pos = 0;
    QTreeWidgetItemIterator it(this);
    while (*it)
    {
        int id = idlist.value(pos);
        if (id == current_songID)
            current_songPos = pos;
        (*it)->set_id(id);
        (*it)->setText(0, QString::number(++pos));
        (*it)->set_pos(pos);
        ++it;
    }
}


void qm_plistview::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Return)
    {
        on_item_dclicked(currentItem());
        return;
    }

    if (event->key() == Qt::Key_Delete)
    {
        delete_it();
        return;
    }

    if (event->key() == Qt::Key_Space)
    {
        if (current_songPos != -1 && current_songPos < topLevelItemCount() )
            scrollToItem(topLevelItem(current_songPos));
        return;
    }

    if (event->key() == Qt::Key_Home)
    {
        if (topLevelItemCount() > 0)
            scrollToItem(topLevelItem(0));
        return;
    }

    if (event->key() == Qt::Key_End)
    {
        if (topLevelItemCount() > 0)
            scrollToItem(topLevelItem(topLevelItemCount() - 1));
        return;
    }

    if (event->key() == Qt::Key_Down)
    {
        if (currentItem()->get_pos < topLevelItemCount())
            setCurrentItem(itemBelow ( currentItem() ));
        return;
    }

    if (event->key() == Qt::Key_Up)
    {
        if (currentItem()->get_pos > 1)
            setCurrentItem(itemAbove ( currentItem() ));
        return;
    }

}


void qm_plistview::delete_it()
{
    if (topLevelItemCount() == 0 || !mpd_connected)
        return;

    QList<QTreeWidgetItem*> tobeRemoved = selectedItems();
    if (tobeRemoved.empty())
        return;

    if (tobeRemoved.size() == topLevelItemCount())
    {
        clear_it();
        return;
    }

    qm_commandList newCommandList;

    QListIterator<QTreeWidgetItem *> itr(tobeRemoved);
    itr.toBack();
    while (itr.hasPrevious())
    {
        QTreeWidgetItem *current = itr.previous();
        qm_mpd_command newCommand;
        newCommand.cmd = CMD_DEL;

        if (current->get_pos < current_songPos)
            current_songPos--;

        int ID = current->get_id;
        newCommand.songid = ID;
        if (ID == current_songID)
        {
            current_songID = -1;
            current_songPos = -1;
        }

        plist_totaltime -=  current->get_time;

        takeTopLevelItem( indexOfTopLevelItem(current) );
        // higher pos first (push_back)
        newCommandList.push_back(newCommand);
    }

    // first
    mpdCom->execute_cmds( newCommandList, false );
    // next
    reNum_reID_list();
    reSize_columns();


    newCommandList.clear();
    tobeRemoved.clear();

    QString plist_stats = trUtf8("<i>Tracks</i> ") + QString::number(topLevelItemCount()) + trUtf8(" <i>- Total time</i> ") + into_time(plist_totaltime);
    emit sgnl_plist_statistics(plist_stats);

    if (current_songID == -1 )
        emit sgnl_curentsong_deleted(); // player will report back and set plist accordingly

}


void qm_plistview::on_item_dclicked(QTreeWidgetItem *item)
{
    int itemID = item->get_id;
    if (itemID != -1 && mpd_connected)
    {
        // a stream may trigger on_mpd_newlist before set_newsong
        // and reset the list. So set ID and b_streamplaying now.
        current_songID = itemID;
        if (item->get_type == TP_STREAM)
            b_streamplaying = true;
        else
            b_streamplaying = false;
        mpdCom->play_this(itemID);
    }
}


void qm_plistview::set_status(int newStatus)
{
    if (current_songPos == -1 || current_songPos >= topLevelItemCount() )
        return;

    switch (newStatus)
    {
    case 1: // MPD_STATE_STOP
        {
            topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_stop.png"));
            if ( topLevelItem(current_songPos)->get_state != STATE_PLAYING )
                for (int i = 0; i < 6; i++)
                    topLevelItem(current_songPos)->setForeground( i, QBrush(col_playing_fg));
            break;
        }

    case 2: // MPD_STATE_PLAY
        {
            topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_play.png"));
            for (int i = 0; i < 6; i++)
                topLevelItem(current_songPos)->setForeground( i, QBrush(col_playing_fg));
            topLevelItem(current_songPos)->set_state(STATE_PLAYING);
            break;
        }

    case 3: // MPD_STATE_PAUSE
        {
            topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_pause.png"));
            if ( topLevelItem(current_songPos)->get_state != STATE_PLAYING )
                for (int i = 0; i < 6; i++)
                    topLevelItem(current_songPos)->setForeground( i, QBrush(col_playing_fg));
            break;
        }

    default: // 0 (MPD_STATE_UNKNOWN) or -1 (Disconnected)
        {
            topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_wait.png"));
            break;
        }
    }
    current_status = newStatus;
}


void qm_plistview::set_newsong(int newPos)
{
    if ( topLevelItemCount() == 0 || !(newPos < topLevelItemCount()) )
        return;

    // reset current
    if (current_songPos != -1)
    {
        switch (topLevelItem(current_songPos)->get_state)
        {

        case STATE_PLAYING:
            {
                if (topLevelItem(current_songPos)->get_type == TP_SONG)
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_trackplayed.png"));
                else
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_streamplayed.png"));
                for (int i = 0; i < 6; i++)
                    topLevelItem(current_songPos)->setForeground( i, QBrush(col_played_fg));
                topLevelItem(current_songPos)->set_state(STATE_PLAYED);
                break;
            }

        case STATE_PLAYED:
            {
                if (topLevelItem(current_songPos)->get_type == TP_SONG)
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_trackplayed.png"));
                else
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_streamplayed.png"));
                for (int i = 0; i < 6; i++)
                    topLevelItem(current_songPos)->setForeground( i, QBrush(col_played_fg));
                break;
            }

        default: // STATE_NEW
            {
                if (topLevelItem(current_songPos)->get_type == TP_SONG)
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_track.png"));
                else
                    topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_stream.png"));
                for (int i = 0; i < 6; i++)
                    topLevelItem(current_songPos)->setForeground( i, QBrush(col_default_fg));
                break;
            }
        }
    }


    if (newPos < 0 ) // 'no song'
    {
        current_songID = -1;
        if (topLevelItemCount() > 0)
        {
            current_songPos = 0;
            set_status(-1);
        }
        else
            current_songPos = -1;
        return;
    }

    current_songID = topLevelItem(newPos)->get_id;

    if (topLevelItem(newPos)->get_type == TP_STREAM)
        b_streamplaying = true;
    else
        b_streamplaying = false;

    current_songPos = newPos;

    set_status(current_status);
}


void qm_plistview::on_mpd_newlist(qm_songList newPlaylist, int changerID)
{
    // prevent resetting by a stream
    if (b_streamplaying)
    {
        // the current song did indeed reset the list
        if (current_songID == changerID)
            return;
    }

    clear();

    QTreeWidgetItem *new_item;
    int pos = 0;
    plist_totaltime = 0;

    QList<qm_songInfo>::iterator iter;
    for (iter = newPlaylist.begin(); iter != newPlaylist.end(); ++iter)
    {
        if (!iter->nosong)
        {
            pos++;
            if (iter->type == TP_STREAM)
            {
                QStringList labels;
                labels << QString::number(pos) << "" << trUtf8(" Stream ") << iter->file << "" << trUtf8(" Info in player ");
                new_item =  new QTreeWidgetItem(labels);
                new_item->set_pos(pos);
                new_item->setIcon(1, QIcon(":/tr_stream.png"));
                new_item->set_id(iter->songID);
                new_item->set_state(STATE_NEW);
                new_item->set_type(TP_STREAM);
                new_item->set_time(0);
                new_item->set_file(iter->file);
                addTopLevelItem(new_item);
            }
            else
                if (iter->type == TP_SONG) // should always be true
                {
                    plist_totaltime += iter->time;
                    if (iter->title.isEmpty())
                    {
                        // use file name as title
                        int i = (iter->file).lastIndexOf( '/' ) + 1;
                        iter->title = (iter->file).right((iter->file).length() - i);
                    }
                    if (iter->album.isEmpty())
                        iter->album = "?";
                    if (iter->artist.isEmpty())
                        iter->artist = "?";
                    if (iter->track.isEmpty())
                        iter->track = "?";
                    QStringList labels;
                    labels  << QString::number(pos)
                    << "" // icon
                    << " " + iter->artist + " "
                    << " " + iter->title + " "
                    << " " + into_time(iter->time) + " "
                    << " " + iter->album + " : " + iter->track + " ";
                    new_item =  new QTreeWidgetItem(labels);
                    new_item->set_pos(pos);
                    new_item->setIcon(1, QIcon(":/tr_track.png"));
                    new_item->set_id(iter->songID);
                    new_item->set_state(STATE_NEW);
                    new_item->set_type(TP_SONG);
                    new_item->set_time(iter->time);
                    new_item->set_file(iter->file);
                    addTopLevelItem(new_item);
                }
        }
    }
    newPlaylist.clear();

    QString plist_stats = trUtf8("<i>Tracks</i> ") + QString::number(topLevelItemCount()) + trUtf8(" <i>- Total time</i> ") + into_time(plist_totaltime);
    emit sgnl_plist_statistics(plist_stats);

    reSize_columns();

    if (topLevelItemCount() > 0)
    {
        current_songPos = 0;
        topLevelItem(current_songPos)->setIcon(1, QIcon(":/tr_wait.png"));
        // Keep current_status! It may be needed in set_newSong.
    }
    else
        current_songPos = -1;
}


QString qm_plistview::into_time(int time)
{
    QString formattedTime = "";
    int hrs, mins, secs;

    hrs = (int) (time / 3600);
    if (hrs > 0)
    {
        formattedTime += QString::number(hrs) + ":";
        mins = (time % 3600)/60;
        if(mins < 10)
            formattedTime += "0";
    }
    else
        mins = (int)(time / 60);

    formattedTime += QString::number(mins) + ":";

    secs = time % 60;
    if(secs < 10)
        formattedTime += "0";
    formattedTime += QString::number(secs);
    return formattedTime;
}


void qm_plistview::select_it(int mode)
{

    QString findThis = config->select_string;

    if (findThis.isEmpty() || topLevelItemCount() == 0)
        return;

    // remove current selections
    QListIterator<QTreeWidgetItem *> itr(selectedItems());
    while (itr.hasNext())
    {
        itr.next()->setSelected(false);
    }

    int column;

    switch (mode)
    {
    case SR_ID_ARTIST:
        {
            column = 2;
            break;
        }

    case SR_ID_TITLE:
        {
            column = 3;
            break;
        }

    case SR_ID_ALBUM:
        {
            column = 5;
            break;
        }

    default: // SR_ID_ALL
        column = 0;
    }

    if (column == 0)
        for(int colm = 2; colm < 6; colm++)
        {
            QList<QTreeWidgetItem *> result = findItems( findThis, Qt::MatchContains|Qt::MatchFixedString, colm );
            QListIterator<QTreeWidgetItem *> it(result);
            while (it.hasNext())
            {
                it.next()->setSelected(true);
            }
            result.clear();
        }
    else
    {
        QList<QTreeWidgetItem *> result = findItems( findThis, Qt::MatchContains|Qt::MatchFixedString, column );
        QListIterator<QTreeWidgetItem *> it(result);
        while (it.hasNext())
        {
            it.next()->setSelected(true);
        }
        result.clear();
    }
}

void qm_plistview::purge_it()
{
    if (topLevelItemCount() == 0 || !mpd_connected )
        return;

    qm_commandList newCommandList;
    QList<QTreeWidgetItem*> tobeRemoved;

    QTreeWidgetItemIterator itr(this);
    while (*itr)
    {
        if ( (*itr)->get_state == STATE_PLAYED)
        {
            qm_mpd_command newCommand;
            newCommand.cmd = CMD_DEL;
            newCommand.songid = (*itr)->get_id;
            // higher pos first (so push_front)
            newCommandList.push_front(newCommand);
            // higher pos last (so push_back)
            tobeRemoved.push_back(*itr);
        }
        ++itr;
    }

    mpdCom->execute_cmds( newCommandList, false );
    newCommandList.clear();

    QListIterator<QTreeWidgetItem *> it(tobeRemoved);
    it.toBack();
    while (it.hasPrevious())
    {
        takeTopLevelItem( indexOfTopLevelItem(it.previous()) );
    }
    tobeRemoved.clear();

    reNum_reID_list();
    reSize_columns();

    QString plist_stats = trUtf8("<i>Tracks</i> ") + QString::number(topLevelItemCount()) + trUtf8(" <i>- Total time</i> ") + into_time(plist_totaltime);
    emit sgnl_plist_statistics(plist_stats);
}


void qm_plistview::clear_it()
{
    if (topLevelItemCount() == 0 || !mpd_connected )
        return;

    clear();
    mpdCom->clear_list();
}


void qm_plistview::shuffle_it()
{
    if (topLevelItemCount() < 2 || !mpd_connected )
        return;

    QTime midnight(0, 0, 0);
    qsrand(midnight.secsTo(QTime::currentTime()));

    QTreeWidgetItemIterator itr(this);
    while (*itr)
    {
        int pos = indexOfTopLevelItem(*itr);
        int moveto = (qrand() % (topLevelItemCount()-1));
        insertTopLevelItem (moveto, takeTopLevelItem(pos) );
        ++itr;
    }

    qm_commandList newCommandList;
    QTreeWidgetItemIterator it(this);
    while (*it)
    {
        qm_mpd_command newCommand;
        newCommand.cmd = CMD_MOV;
        newCommand.songid = (*it)->get_id;
        newCommand.moveTo = indexOfTopLevelItem(*it);
        if ( newCommand.songid == current_songID )
            current_songPos = newCommand.moveTo;
        newCommandList.push_back(newCommand);
        ++it;
    }
    mpdCom->execute_cmds( newCommandList, false );
    newCommandList.clear();

    // move current to pos o
    if (current_songPos > 0)
    {
        insertTopLevelItem (0, takeTopLevelItem(current_songPos) );
        current_songPos = 0;
        qm_mpd_command newCommand;
        newCommand.cmd = CMD_MOV;
        newCommand.songid = current_songID;
        newCommand.moveTo = 0;
        newCommandList.push_back(newCommand);
        mpdCom->execute_cmds( newCommandList, false );
        newCommandList.clear();
    }
    reNum_reID_list();
}


void qm_plistview::save_it()
{
    if (topLevelItemCount() == 0 )
        return;

    QString suggested_name = (topLevelItem(0)->text(2)).simplified();
    bool ok;
    QString listname = QInputDialog::getText(this, trUtf8("Save"),
                       trUtf8(" Enter a name for the playlist "), QLineEdit::Normal,
                       suggested_name, &ok);
    if ( !ok || listname.isEmpty() )
        return;

    QString result = mpdCom->save_playlist(listname);

    if (result.startsWith("Error:"))
    {
        result = result.section("} ",1);
        QMessageBox::warning( this, trUtf8("Error"), result);
    }
    else
        QMessageBox::information( this, trUtf8("Confirmation"), result);
}


void qm_plistview::save_selected()
{
    if (topLevelItemCount() == 0 )
        return;

    QList<QTreeWidgetItem*> selected = selectedItems();
    if (selected.empty())
        return;

    if (selected.size() == topLevelItemCount())
    {
        save_it();
        return;
    }

    bool ok;
    QString listname = QInputDialog::getText(this, trUtf8("Save"),
                       trUtf8(" Enter a name for the playlist "), QLineEdit::Normal,
                       "selection", &ok);
    if ( !ok || listname.isEmpty() )
        return;

    QStringList tobeSaved;

    QListIterator<QTreeWidgetItem *> itr(selected);
    while (itr.hasNext())
    {
        tobeSaved.push_back(itr.next()->get_file);
    }

    QString result = mpdCom->save_selection(tobeSaved, listname);

    if (result.startsWith("Error:"))
    {
        result = result.section("} ",1);
        QMessageBox::warning( this, trUtf8("Error"), result);
    }
    else
        QMessageBox::information( this, trUtf8("Confirmation"), result);
}


void qm_plistview::set_font(QFont & thefont)
{
    setFont(thefont);
    reSize_columns();
}


void qm_plistview::set_mpdCom(qm_mpdCom* conn, bool isconnected)
{
    mpd_connected = isconnected;
    mpdCom = conn;
    QObject::connect( mpdCom, SIGNAL( sgnl_plistupdate (qm_songList, int)), this, SLOT( on_mpd_newlist(qm_songList, int)));

    if (mpd_connected && config->mpd_save_allowed)
    {
        a_savelist->setEnabled(true);
        a_savesel->setEnabled(true);
    }
    else
    {
        a_savelist->setEnabled(false);
        a_savesel->setEnabled(false);
    }

    if (config->mpd_shuffle_allowed)
        a_shuffle->setEnabled(true);
    else
        a_shuffle->setEnabled(false);

    if (config->mpd_deleteid_allowed)
        a_delsel->setEnabled(true);
    else
        a_delsel->setEnabled(false);
}


void qm_plistview::contextMenuEvent ( QContextMenuEvent * event )
{
    if (event->reason() == QContextMenuEvent::Mouse && mpd_connected)
        ctxMenu->exec(QCursor::pos());
}


void qm_plistview::set_config(qm_config *conf)
{
    if (conf == NULL)
        return;

    config = conf;
}


/*  Pointer to thePlaylistModList in browserWin.
 plist_view and lib_view share this list for playlist
 modifications like drag&drop operations. */
void qm_plistview::set_playlistModList(qm_commandList* theList)
{
    thePlaylistModList = theList;
}

void qm_plistview::on_playlist_modification(bool dragdrop)
{
    bool newlist = false;
    int pos = topLevelItemCount();
    if (pos == 0)
    {
        plist_totaltime = 0;
        newlist = true;
    }
    if (dragdrop)
        pos = plist_droppos;

    QTreeWidgetItem *new_item;
    QListIterator<qm_mpd_command> itr(*thePlaylistModList);
    while (itr.hasNext())
    {
        qm_mpd_command curCommand = itr.next();
        if ( curCommand.type == TP_STREAM)
        {
            QStringList labels;
            labels << QString::number(++pos) << "" << trUtf8(" Stream ") << curCommand.file << "" << trUtf8(" Info in player ");
            new_item =  new QTreeWidgetItem(labels);
            new_item->setIcon(1, QIcon(":/tr_stream.png"));
            new_item->set_state(STATE_NEW);
            new_item->set_type(TP_STREAM);
            new_item->set_time(0);
            new_item->set_file(curCommand.file);
            if (curCommand.cmd == CMD_ADD)
                addTopLevelItem(new_item);
            else
                insertTopLevelItem(curCommand.moveTo, new_item);
        }
        else
            if ( curCommand.type == TP_SONG) // should always be true
            {
                plist_totaltime += curCommand.time;
                if (curCommand.title.isEmpty())
                {
                    // use file name as title
                    int i = (curCommand.file).lastIndexOf( '/' ) + 1;
                    curCommand.title = (curCommand.file).right((curCommand.file).length() - i);
                }
                if (curCommand.album.isEmpty())
                    curCommand.album = "?";
                if (curCommand.artist.isEmpty())
                    curCommand.artist = "?";
                if (curCommand.track.isEmpty())
                    curCommand.track = "?";
                QStringList labels;
                labels  << QString::number(++pos)
                << ""  // icon
                << " " + curCommand.artist + " "
                << " " + curCommand.title + " "
                << " " + into_time(curCommand.time) + " "
                << " " + curCommand.album + " : " + curCommand.track + " ";
                new_item =  new QTreeWidgetItem(labels);
                new_item->setIcon(1, QIcon(":/tr_track.png"));
                new_item->set_state(STATE_NEW);
                new_item->set_type(TP_SONG);
                new_item->set_time(curCommand.time);
                new_item->set_file(curCommand.file);
                if (curCommand.cmd == CMD_ADD)
                    addTopLevelItem(new_item);
                else
                    insertTopLevelItem(curCommand.moveTo, new_item);
            }
    }

    // if we created a new list
    if (newlist && topLevelItemCount() > 0)
    {
        current_songPos = 0;
        topLevelItem(0)->setIcon(1, QIcon(":/tr_wait.png"));
    }

    reNum_reID_list();
    reSize_columns();

    QString plist_stats = trUtf8("<i>Tracks</i> ") + QString::number(topLevelItemCount()) + trUtf8(" <i>- Total time</i> ") + into_time(plist_totaltime);
    emit sgnl_plist_statistics(plist_stats);
}


void qm_plistview::startDrag(Qt::DropActions da)
{
    QMimeData *data;
    QModelIndexList indexes = selectedIndexes();
    if (indexes.count() > 0)
    {
        data = model()->mimeData(indexes);
        if (data == 0)
            return;
    }
    else
        return;
    QDrag* drag = new QDrag(this);
    drag->setPixmap(QPixmap(":/tr_dragdrop.png"));
    drag->setMimeData(data);
    drag->exec(da, Qt::IgnoreAction);
}


void qm_plistview::dragMoveEvent(QDragMoveEvent *e)
{
    if (e->source() == 0)
        e->ignore();
    else
    {
        QTreeWidgetItem *item = itemAt(e->pos());
        if (item)
        {
            dropBeforeIndex = indexOfTopLevelItem(item);
            showLine_at(visualItemRect(item).top());
        }
        else if (e->pos().y() >= 0)
        {
            dropBeforeIndex = topLevelItemCount();
            showLine_at(visualItemRect(topLevelItem(topLevelItemCount() - 1)).bottom());
        }
        else
        {
            dropBeforeIndex = -1;
            line->hide();
        }
        e->accept();
    }
}

void qm_plistview::dragEnterEvent(QDragEnterEvent *e) ///////////////////////////// also works on drag from self!!!
{
    if (e->source() > 0) // source within application
    {
        line->show();
        e->accept();
    }
    else
        e->ignore();
}

void qm_plistview::dragLeaveEvent(QDragLeaveEvent *e) ///////////////////////////// also works on drag from self!!! // b_drag_from_self stays TRUE after leave event!
{
    line->hide();
    dropBeforeIndex = -1;
    e->accept();
}

void qm_plistview::dropEvent(QDropEvent *e)
{
    if (e->source() == this) // drag from inside playlist
    {
        QList<QTreeWidgetItem*> tobeMoved = selectedItems();

        if (topLevelItemCount() == 0 || !mpd_connected || tobeMoved.empty() || tobeMoved.size() == topLevelItemCount())
        {
            this->setFocus();
            line->hide();
            dropBeforeIndex = -1;
            e->ignore();
            return;
        }

        // first rearrange the treeview
        QListIterator<QTreeWidgetItem *> itr(tobeMoved);
        while (itr.hasNext())
        {
            QTreeWidgetItem *current = itr.next();

            if (current->get_pos < dropBeforeIndex)
            {
                QTreeWidgetItem *taken = takeTopLevelItem( indexOfTopLevelItem(current) );
                insertTopLevelItem ( dropBeforeIndex -1, taken );
            }
            else
            {
                QTreeWidgetItem *taken = takeTopLevelItem( indexOfTopLevelItem(current) );
                insertTopLevelItem ( dropBeforeIndex, taken );
            }
        }

        // next sync the mpd playlist with the treeview
        int pos = 0;
        qm_commandList newCommandList;
        QTreeWidgetItemIterator it(this);
        while (*it)
        {
            qm_mpd_command newCommand;
            newCommand.cmd = CMD_MOV;
            newCommand.songid = (*it)->get_id;
            newCommand.moveTo = pos;
            newCommandList.push_back(newCommand);
            pos++;
            ++it;
        }

        mpdCom->execute_cmds( newCommandList, false );
        reNum_reID_list();
        reSize_columns();

        newCommandList.clear();
        tobeMoved.clear();

        this->setFocus();
        line->hide();
        dropBeforeIndex = -1;
        e->ignore();
    }

    if (e->source() != this && e->source() != 0) // drag from library
    {
        if (topLevelItemCount() == 0)
        {
            plist_droppos = -1;
            emit sgnl_plist_dropreceived(plist_droppos);
        }
        else
            if (dropBeforeIndex != -1)
            {
                plist_droppos = dropBeforeIndex;
                emit sgnl_plist_dropreceived(plist_droppos);
            }

        this->setFocus();
        line->hide();
        dropBeforeIndex = -1;

        e->ignore();
        return;
    }

    if (e->source() == 0) // drag from outside the app
    {
        e->ignore();
        return;
    }
}

void qm_plistview::showLine_at(int y)
{
    line->setGeometry(2, y, viewport()->width()-4, 2);
}

qm_plistview::~qm_plistview()
{}
